import asyncio
import json
import time
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Dict,
    Generator,
    Generic,
    List,
    Optional,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
)

import sqlalchemy as sa

from dbos._context import MaxPriority, MinPriority
from dbos._core import DEFAULT_POLLING_INTERVAL
from dbos._sys_db import SystemDatabase
from dbos._utils import generate_uuid

if TYPE_CHECKING:
    from dbos._dbos import WorkflowHandle, WorkflowHandleAsync

from dbos._croniter import croniter  # type: ignore
from dbos._dbos_config import get_system_database_url, is_valid_database_url
from dbos._error import DBOSException, DBOSNonExistentWorkflowError
from dbos._registrations import DEFAULT_MAX_RECOVERY_ATTEMPTS
from dbos._scheduler import backfill_schedule, trigger_schedule
from dbos._serialization import (
    DefaultSerializer,
    Serializer,
    WorkflowSerializationFormat,
    serialize_args,
)
from dbos._sys_db import (
    ClientScheduleInput,
    StepInfo,
    SystemDatabase,
    VersionInfo,
    WorkflowSchedule,
    WorkflowStatus,
    WorkflowStatusInternal,
    WorkflowStatusString,
    _dbos_stream_closed_sentinel,
    workflow_is_active,
)
from dbos._workflow_commands import fork_workflow, get_workflow

R = TypeVar("R", covariant=True)  # A generic type for workflow return values


# Required EnqueueOptions fields
class _EnqueueOptionsRequired(TypedDict):
    workflow_name: str
    queue_name: str


# Optional EnqueueOptions fields
class EnqueueOptions(_EnqueueOptionsRequired, total=False):
    workflow_id: str
    app_version: str
    workflow_timeout: float
    deduplication_id: str
    priority: int
    max_recovery_attempts: int
    queue_partition_key: str
    authenticated_user: str
    authenticated_roles: list[str]
    serialization_type: WorkflowSerializationFormat
    class_name: str
    instance_name: str


def validate_enqueue_options(options: EnqueueOptions) -> None:
    priority = options.get("priority")
    if priority is not None and (priority < MinPriority or priority > MaxPriority):
        raise DBOSException(
            f"Invalid priority {priority}. Priority must be between {MinPriority}~{MaxPriority}."
        )


class WorkflowHandleClientPolling(Generic[R]):

    def __init__(self, workflow_id: str, sys_db: SystemDatabase):
        self.workflow_id = workflow_id
        self._sys_db = sys_db

    def get_workflow_id(self) -> str:
        return self.workflow_id

    def get_result(
        self, *, polling_interval_sec: float = DEFAULT_POLLING_INTERVAL
    ) -> R:
        res: R = self._sys_db.await_workflow_result(
            self.workflow_id, polling_interval_sec
        )
        return res

    def get_status(self) -> WorkflowStatus:
        status = get_workflow(self._sys_db, self.workflow_id)
        if status is None:
            raise DBOSNonExistentWorkflowError("target", self.workflow_id)
        return status


class WorkflowHandleClientAsyncPolling(Generic[R]):

    def __init__(self, workflow_id: str, sys_db: SystemDatabase):
        self.workflow_id = workflow_id
        self._sys_db = sys_db

    def get_workflow_id(self) -> str:
        return self.workflow_id

    async def get_result(
        self, *, polling_interval_sec: float = DEFAULT_POLLING_INTERVAL
    ) -> R:
        res: R = await self._sys_db.await_workflow_result_async(
            self.workflow_id, polling_interval_sec
        )
        return res

    async def get_status(self) -> WorkflowStatus:
        status = await asyncio.to_thread(get_workflow, self._sys_db, self.workflow_id)
        if status is None:
            raise DBOSNonExistentWorkflowError("target", self.workflow_id)
        return status


class DBOSClient:

    def __init__(
        self,
        database_url: Optional[str] = None,  # DEPRECATED
        *,
        system_database_url: Optional[str] = None,
        system_database_engine: Optional[sa.Engine] = None,
        application_database_url: Optional[str] = None,
        dbos_system_schema: Optional[str] = "dbos",
        serializer: Serializer = DefaultSerializer(),
    ):
        self._serializer = serializer
        if system_database_engine:
            if "sqlite" in system_database_engine.dialect.name:
                system_database_url = "sqlite:///custom_system_database_engine.sqlite"
            else:
                system_database_url = "postgresql://custom:system@database/engine"
        else:
            application_database_url = (
                database_url if database_url else application_database_url
            )
            system_database_url = get_system_database_url(
                {
                    "system_database_url": system_database_url,
                    "database_url": application_database_url,
                }
            )
            assert is_valid_database_url(system_database_url)
        # We only create database connections but do not run migrations
        self._sys_db = SystemDatabase.create(
            system_database_url=system_database_url,
            engine_kwargs={
                "connect_args": {"application_name": "dbos_transact_client"},
                "pool_timeout": 30,
                "max_overflow": 0,
                "pool_size": 2,
                "pool_pre_ping": True,
            },
            engine=system_database_engine,
            schema=dbos_system_schema,
            serializer=serializer,
            executor_id=None,
        )
        self._sys_db.check_connection()

    def destroy(self) -> None:
        self._sys_db.destroy()

    def _enqueue(self, options: EnqueueOptions, *args: Any, **kwargs: Any) -> str:
        validate_enqueue_options(options)
        workflow_name = options["workflow_name"]
        queue_name = options["queue_name"]

        max_recovery_attempts = options.get("max_recovery_attempts")
        if max_recovery_attempts is None:
            max_recovery_attempts = DEFAULT_MAX_RECOVERY_ATTEMPTS
        workflow_id = options.get("workflow_id")
        if workflow_id is None:
            workflow_id = generate_uuid()
        workflow_timeout = options.get("workflow_timeout", None)

        authenticated_user = options.get("authenticated_user")
        authenticated_roles = (
            json.dumps(options.get("authenticated_roles"))
            if options.get("authenticated_roles")
            else None
        )

        inputs, serialization = serialize_args(
            args,
            kwargs,
            options.get("serialization_type"),
            self._serializer,
        )

        status: WorkflowStatusInternal = {
            "workflow_uuid": workflow_id,
            "status": WorkflowStatusString.ENQUEUED.value,
            "name": workflow_name,
            "class_name": options.get("class_name"),
            "queue_name": queue_name,
            "app_version": options.get("app_version"),
            "config_name": options.get("instance_name"),
            "authenticated_user": authenticated_user,
            "assumed_role": None,
            "authenticated_roles": authenticated_roles,
            "output": None,
            "error": None,
            "created_at": None,
            "updated_at": None,
            "executor_id": None,
            "recovery_attempts": None,
            "app_id": None,
            "workflow_timeout_ms": (
                int(workflow_timeout * 1000) if workflow_timeout is not None else None
            ),
            "workflow_deadline_epoch_ms": None,
            "deduplication_id": options.get("deduplication_id", None),
            "priority": (
                options.get("priority", 0)
                if options.get("priority", None) is not None
                else 0
            ),
            "inputs": inputs,
            "serialization": serialization,
            "queue_partition_key": options.get("queue_partition_key", None),
            "forked_from": None,
            "parent_workflow_id": None,
            "started_at_epoch_ms": None,
            "owner_xid": None,
        }

        self._sys_db.init_workflow(
            status,
            max_recovery_attempts=None,
            owner_xid=None,
            is_dequeued_request=False,
            is_recovery_request=False,
        )
        return workflow_id

    def enqueue(
        self, options: EnqueueOptions, *args: Any, **kwargs: Any
    ) -> "WorkflowHandle[R]":
        workflow_id = self._enqueue(options, *args, **kwargs)
        return WorkflowHandleClientPolling[R](workflow_id, self._sys_db)

    async def enqueue_async(
        self, options: EnqueueOptions, *args: Any, **kwargs: Any
    ) -> "WorkflowHandleAsync[R]":
        workflow_id = await asyncio.to_thread(self._enqueue, options, *args, **kwargs)
        return WorkflowHandleClientAsyncPolling[R](workflow_id, self._sys_db)

    def retrieve_workflow(self, workflow_id: str) -> "WorkflowHandle[R]":
        status = get_workflow(self._sys_db, workflow_id)
        if status is None:
            raise DBOSNonExistentWorkflowError("target", workflow_id)
        return WorkflowHandleClientPolling[R](workflow_id, self._sys_db)

    async def retrieve_workflow_async(
        self, workflow_id: str
    ) -> "WorkflowHandleAsync[R]":
        status = await asyncio.to_thread(get_workflow, self._sys_db, workflow_id)
        if status is None:
            raise DBOSNonExistentWorkflowError("target", workflow_id)
        return WorkflowHandleClientAsyncPolling[R](workflow_id, self._sys_db)

    def wait_first(
        self,
        handles: List["WorkflowHandle[Any]"],
        *,
        polling_interval_sec: float = DEFAULT_POLLING_INTERVAL,
    ) -> "WorkflowHandle[Any]":
        """Wait for any one of the given workflow handles to complete and return it."""
        if not handles:
            raise ValueError("handles must not be empty")
        workflow_ids = [h.workflow_id for h in handles]
        if len(set(workflow_ids)) != len(workflow_ids):
            raise ValueError("handles must not contain duplicate workflow IDs")
        handle_map: Dict[str, "WorkflowHandle[Any]"] = {
            h.workflow_id: h for h in handles
        }
        completed_id = self._sys_db.await_first_workflow_id(
            workflow_ids, polling_interval_sec
        )
        return handle_map[completed_id]

    async def wait_first_async(
        self,
        handles: List["WorkflowHandleAsync[Any]"],
        *,
        polling_interval_sec: float = DEFAULT_POLLING_INTERVAL,
    ) -> "WorkflowHandleAsync[Any]":
        """Async version of :meth:`wait_first`."""
        if not handles:
            raise ValueError("handles must not be empty")
        workflow_ids = [h.workflow_id for h in handles]
        if len(set(workflow_ids)) != len(workflow_ids):
            raise ValueError("handles must not contain duplicate workflow IDs")
        handle_map: Dict[str, "WorkflowHandleAsync[Any]"] = {
            h.workflow_id: h for h in handles
        }
        completed_id = await asyncio.to_thread(
            self._sys_db.await_first_workflow_id,
            workflow_ids,
            polling_interval_sec,
        )
        return handle_map[completed_id]

    def send(
        self,
        destination_id: str,
        message: Any,
        topic: Optional[str] = None,
        idempotency_key: Optional[str] = None,
        *,
        serialization_type: Optional[
            WorkflowSerializationFormat
        ] = WorkflowSerializationFormat.DEFAULT,
    ) -> None:
        self._sys_db.send_direct(
            destination_id,
            message,
            topic,
            message_uuid=idempotency_key,
            serialization_type=serialization_type,
        )

    async def send_async(
        self,
        destination_id: str,
        message: Any,
        topic: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> None:
        return await asyncio.to_thread(
            self.send, destination_id, message, topic, idempotency_key
        )

    def get_event(self, workflow_id: str, key: str, timeout_seconds: float = 60) -> Any:
        return self._sys_db.get_event(workflow_id, key, timeout_seconds)

    async def get_event_async(
        self, workflow_id: str, key: str, timeout_seconds: float = 60
    ) -> Any:
        return await asyncio.to_thread(
            self.get_event, workflow_id, key, timeout_seconds
        )

    def cancel_workflow(self, workflow_id: str) -> None:
        self._sys_db.cancel_workflow(workflow_id)

    async def cancel_workflow_async(self, workflow_id: str) -> None:
        await asyncio.to_thread(self.cancel_workflow, workflow_id)

    def resume_workflow(self, workflow_id: str) -> "WorkflowHandle[Any]":
        self._sys_db.resume_workflow(workflow_id)
        return WorkflowHandleClientPolling[Any](workflow_id, self._sys_db)

    async def resume_workflow_async(
        self, workflow_id: str
    ) -> "WorkflowHandleAsync[Any]":
        await asyncio.to_thread(self.resume_workflow, workflow_id)
        return WorkflowHandleClientAsyncPolling[Any](workflow_id, self._sys_db)

    def list_workflows(
        self,
        *,
        workflow_ids: Optional[List[str]] = None,
        status: Optional[str | list[str]] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        name: Optional[str | list[str]] = None,
        app_version: Optional[str | list[str]] = None,
        forked_from: Optional[str | list[str]] = None,
        parent_workflow_id: Optional[str | list[str]] = None,
        user: Optional[str | list[str]] = None,
        queue_name: Optional[str | list[str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort_desc: bool = False,
        workflow_id_prefix: Optional[str | list[str]] = None,
        load_input: bool = True,
        load_output: bool = True,
        executor_id: Optional[str | list[str]] = None,
        queues_only: bool = False,
    ) -> List[WorkflowStatus]:
        return self._sys_db.list_workflows(
            workflow_ids=workflow_ids,
            status=status,
            start_time=start_time,
            end_time=end_time,
            name=name,
            app_version=app_version,
            forked_from=forked_from,
            parent_workflow_id=parent_workflow_id,
            user=user,
            queue_name=queue_name,
            limit=limit,
            offset=offset,
            sort_desc=sort_desc,
            workflow_id_prefix=workflow_id_prefix,
            load_input=load_input,
            load_output=load_output,
            executor_id=executor_id,
            queues_only=queues_only,
        )

    async def list_workflows_async(
        self,
        *,
        workflow_ids: Optional[List[str]] = None,
        status: Optional[str | list[str]] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        name: Optional[str | list[str]] = None,
        app_version: Optional[str | list[str]] = None,
        forked_from: Optional[str | list[str]] = None,
        parent_workflow_id: Optional[str | list[str]] = None,
        user: Optional[str | list[str]] = None,
        queue_name: Optional[str | list[str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort_desc: bool = False,
        workflow_id_prefix: Optional[str | list[str]] = None,
        load_input: bool = True,
        load_output: bool = True,
        executor_id: Optional[str | list[str]] = None,
        queues_only: bool = False,
    ) -> List[WorkflowStatus]:
        return await asyncio.to_thread(
            self.list_workflows,
            workflow_ids=workflow_ids,
            status=status,
            start_time=start_time,
            end_time=end_time,
            name=name,
            app_version=app_version,
            forked_from=forked_from,
            parent_workflow_id=parent_workflow_id,
            user=user,
            queue_name=queue_name,
            limit=limit,
            offset=offset,
            sort_desc=sort_desc,
            workflow_id_prefix=workflow_id_prefix,
            load_input=load_input,
            load_output=load_output,
            executor_id=executor_id,
            queues_only=queues_only,
        )

    def list_queued_workflows(
        self,
        *,
        workflow_ids: Optional[List[str]] = None,
        status: Optional[str | list[str]] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        name: Optional[str | list[str]] = None,
        app_version: Optional[str | list[str]] = None,
        forked_from: Optional[str | list[str]] = None,
        parent_workflow_id: Optional[str | list[str]] = None,
        user: Optional[str | list[str]] = None,
        queue_name: Optional[str | list[str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort_desc: bool = False,
        workflow_id_prefix: Optional[str | list[str]] = None,
        load_input: bool = True,
        load_output: bool = True,
        executor_id: Optional[str | list[str]] = None,
    ) -> List[WorkflowStatus]:
        return self._sys_db.list_workflows(
            workflow_ids=workflow_ids,
            status=status,
            start_time=start_time,
            end_time=end_time,
            name=name,
            app_version=app_version,
            forked_from=forked_from,
            parent_workflow_id=parent_workflow_id,
            user=user,
            queue_name=queue_name,
            limit=limit,
            offset=offset,
            sort_desc=sort_desc,
            workflow_id_prefix=workflow_id_prefix,
            load_input=load_input,
            load_output=load_output,
            executor_id=executor_id,
            queues_only=True,
        )

    async def list_queued_workflows_async(
        self,
        *,
        workflow_ids: Optional[List[str]] = None,
        status: Optional[str | list[str]] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        name: Optional[str | list[str]] = None,
        app_version: Optional[str | list[str]] = None,
        forked_from: Optional[str | list[str]] = None,
        parent_workflow_id: Optional[str | list[str]] = None,
        user: Optional[str | list[str]] = None,
        queue_name: Optional[str | list[str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort_desc: bool = False,
        workflow_id_prefix: Optional[str | list[str]] = None,
        load_input: bool = True,
        load_output: bool = True,
        executor_id: Optional[str | list[str]] = None,
    ) -> List[WorkflowStatus]:
        return await asyncio.to_thread(
            self.list_queued_workflows,
            workflow_ids=workflow_ids,
            status=status,
            start_time=start_time,
            end_time=end_time,
            name=name,
            app_version=app_version,
            forked_from=forked_from,
            parent_workflow_id=parent_workflow_id,
            user=user,
            queue_name=queue_name,
            limit=limit,
            offset=offset,
            sort_desc=sort_desc,
            workflow_id_prefix=workflow_id_prefix,
            load_input=load_input,
            load_output=load_output,
            executor_id=executor_id,
        )

    def list_workflow_steps(self, workflow_id: str) -> List[StepInfo]:
        return self._sys_db.list_workflow_steps(workflow_id)

    async def list_workflow_steps_async(self, workflow_id: str) -> List[StepInfo]:
        return await asyncio.to_thread(self.list_workflow_steps, workflow_id)

    def fork_workflow(
        self,
        workflow_id: str,
        start_step: int,
        *,
        application_version: Optional[str] = None,
    ) -> "WorkflowHandle[Any]":
        forked_workflow_id = fork_workflow(
            self._sys_db,
            workflow_id,
            start_step,
            application_version=application_version,
        )
        return WorkflowHandleClientPolling[Any](forked_workflow_id, self._sys_db)

    async def fork_workflow_async(
        self,
        workflow_id: str,
        start_step: int,
        *,
        application_version: Optional[str] = None,
    ) -> "WorkflowHandleAsync[Any]":
        forked_workflow_id = await asyncio.to_thread(
            fork_workflow,
            self._sys_db,
            workflow_id,
            start_step,
            application_version=application_version,
        )
        return WorkflowHandleClientAsyncPolling[Any](forked_workflow_id, self._sys_db)

    def read_stream(self, workflow_id: str, key: str) -> Generator[Any, Any, None]:
        """
        Read values from a stream as a generator.
        This function reads values from a stream identified by the workflow_id and key,
        yielding each value in order until the stream is closed or the workflow terminates.

        Args:
            workflow_id: The ID of the workflow that wrote to the stream
            key: The stream key to read from

        Yields:
            The values written to the stream in order
        """
        offset = 0
        while True:
            try:
                value = self._sys_db.read_stream(workflow_id, key, offset)
                if value == _dbos_stream_closed_sentinel:
                    break
                yield value
                offset += 1
            except ValueError:
                # Poll the offset until a value arrives or the workflow terminates
                status = get_workflow(self._sys_db, workflow_id)
                if status is None:
                    break
                if not workflow_is_active(status.status):
                    break
                time.sleep(1.0)
                continue

    async def read_stream_async(
        self, workflow_id: str, key: str
    ) -> AsyncGenerator[Any, None]:
        """
        Read values from a stream as an async generator.
        This function reads values from a stream identified by the workflow_id and key,
        yielding each value in order until the stream is closed or the workflow terminates.

        Args:
            workflow_id: The ID of the workflow that wrote to the stream
            key: The stream key to read from

        Yields:
            The values written to the stream in order
        """
        offset = 0
        while True:
            try:
                value = await asyncio.to_thread(
                    self._sys_db.read_stream, workflow_id, key, offset
                )
                if value == _dbos_stream_closed_sentinel:
                    break
                yield value
                offset += 1
            except ValueError:
                # Poll the offset until a value arrives or the workflow terminates
                status = await asyncio.to_thread(
                    get_workflow, self._sys_db, workflow_id
                )
                if status is None:
                    break
                if not workflow_is_active(status.status):
                    break
                await asyncio.sleep(1.0)
                continue

    # ── Schedule API ──────────────────────────────────────────────

    def create_schedule(
        self,
        *,
        schedule_name: str,
        workflow_name: str,
        schedule: str,
        context: Any = None,
        workflow_class_name: Optional[str] = None,
    ) -> None:
        """
        Create a cron schedule that periodically invokes a workflow.

        Args:
            schedule_name: Unique name identifying this schedule.
            workflow_name: Fully-qualified name of the workflow function to invoke.
            schedule: A cron expression (supports seconds with 6 fields).
            context: A context object passed as the second argument to every invocation. Defaults to ``None``.
            workflow_class_name: Class name for static class method workflows. Defaults to ``None``.

        Raises:
            DBOSException: If the cron expression is invalid or a schedule with the same name already exists.
        """
        if not croniter.is_valid(schedule, second_at_beginning=True):
            raise DBOSException(f"Invalid cron schedule: '{schedule}'")
        self._sys_db.create_schedule(
            WorkflowSchedule(
                schedule_id=generate_uuid(),
                schedule_name=schedule_name,
                workflow_name=workflow_name,
                workflow_class_name=workflow_class_name,
                schedule=schedule,
                status="ACTIVE",
                context=self._sys_db.serializer.serialize(context),
            )
        )

    def list_schedules(
        self,
        *,
        status: Optional[Union[str, List[str]]] = None,
        workflow_name: Optional[Union[str, List[str]]] = None,
        schedule_name_prefix: Optional[Union[str, List[str]]] = None,
    ) -> List[WorkflowSchedule]:
        """
        Return all registered workflow schedules, optionally filtered.

        Args:
            status: Filter by status (e.g. ``"ACTIVE"``) or a list of statuses
            workflow_name: Filter by workflow name or a list of names
            schedule_name_prefix: Filter by schedule name prefix or a list of prefixes
        """
        schedules = self._sys_db.list_schedules(
            status=status,
            workflow_name=workflow_name,
            schedule_name_prefix=schedule_name_prefix,
        )
        for s in schedules:
            s["context"] = self._sys_db.serializer.deserialize(s["context"])
        return schedules

    def get_schedule(self, name: str) -> Optional[WorkflowSchedule]:
        """Return the schedule with the given name, or ``None`` if it does not exist."""
        schedule = self._sys_db.get_schedule(name)
        if schedule is not None:
            schedule["context"] = self._sys_db.serializer.deserialize(
                schedule["context"]
            )
        return schedule

    def delete_schedule(self, name: str) -> None:
        """Delete the schedule with the given name. No-op if it does not exist."""
        self._sys_db.delete_schedule(name)

    async def create_schedule_async(
        self,
        *,
        schedule_name: str,
        workflow_name: str,
        schedule: str,
        context: Any = None,
        workflow_class_name: Optional[str] = None,
    ) -> None:
        """Async version of :meth:`create_schedule`."""
        await asyncio.to_thread(
            self.create_schedule,
            schedule_name=schedule_name,
            workflow_name=workflow_name,
            schedule=schedule,
            context=context,
            workflow_class_name=workflow_class_name,
        )

    async def list_schedules_async(
        self,
        *,
        status: Optional[Union[str, List[str]]] = None,
        workflow_name: Optional[Union[str, List[str]]] = None,
        schedule_name_prefix: Optional[Union[str, List[str]]] = None,
    ) -> List[WorkflowSchedule]:
        """Async version of :meth:`list_schedules`."""
        return await asyncio.to_thread(
            self.list_schedules,
            status=status,
            workflow_name=workflow_name,
            schedule_name_prefix=schedule_name_prefix,
        )

    async def get_schedule_async(self, name: str) -> Optional[WorkflowSchedule]:
        """Async version of :meth:`get_schedule`."""
        return await asyncio.to_thread(self.get_schedule, name)

    async def delete_schedule_async(self, name: str) -> None:
        """Async version of :meth:`delete_schedule`."""
        await asyncio.to_thread(self.delete_schedule, name)

    def pause_schedule(self, name: str) -> None:
        """Pause the schedule with the given name. A paused schedule does not fire."""
        self._sys_db.pause_schedule(name)

    def resume_schedule(self, name: str) -> None:
        """Resume a paused schedule so it begins firing again."""
        self._sys_db.resume_schedule(name)

    def apply_schedules(
        self,
        schedules: List[ClientScheduleInput],
    ) -> None:
        """
        Atomically create or replace a set of schedules.

        Args:
            schedules: A list of schedule inputs, each containing ``schedule_name``, ``workflow_name``, ``schedule`` (cron), and ``context``.

        Raises:
            DBOSException: If a cron expression is invalid
        """
        to_apply: List[WorkflowSchedule] = []
        for entry in schedules:
            cron = entry["schedule"]
            if not croniter.is_valid(cron, second_at_beginning=True):
                raise DBOSException(f"Invalid cron schedule: '{cron}'")
            to_apply.append(
                WorkflowSchedule(
                    schedule_id=generate_uuid(),
                    schedule_name=entry["schedule_name"],
                    workflow_name=entry["workflow_name"],
                    workflow_class_name=entry.get("workflow_class_name"),
                    schedule=cron,
                    status="ACTIVE",
                    context=self._sys_db.serializer.serialize(entry["context"]),
                )
            )
        with self._sys_db.engine.begin() as c:
            for sched in to_apply:
                self._sys_db.delete_schedule(sched["schedule_name"], conn=c)
                self._sys_db.create_schedule(sched, conn=c)

    def backfill_schedule(
        self, schedule_name: str, start: datetime, end: datetime
    ) -> List["WorkflowHandle[None]"]:
        """
        Enqueue all executions of a schedule that would have run between ``start`` and ``end``.

        Each execution uses the same deterministic workflow ID as the live scheduler,
        so already-executed times are skipped.

        Args:
            schedule_name: Name of an existing schedule
            start: Start of the backfill window (exclusive)
            end: End of the backfill window (exclusive)

        Returns:
            A list of workflow handles for each enqueued execution.

        Raises:
            DBOSException: If the schedule does not exist
        """
        workflow_ids = backfill_schedule(self._sys_db, schedule_name, start, end)
        return [
            WorkflowHandleClientPolling[None](wf_id, self._sys_db)
            for wf_id in workflow_ids
        ]

    def trigger_schedule(self, schedule_name: str) -> "WorkflowHandle[None]":
        """
        Immediately enqueue the scheduled workflow at the current time.

        Args:
            schedule_name: Name of an existing schedule

        Returns:
            A workflow handle for the enqueued execution.

        Raises:
            DBOSException: If the schedule does not exist
        """
        workflow_id = trigger_schedule(self._sys_db, schedule_name)
        return WorkflowHandleClientPolling[None](workflow_id, self._sys_db)

    # ── Application Version API ─────────────────────────────────

    def list_application_versions(self) -> List[VersionInfo]:
        """Return all application versions ordered by timestamp descending."""
        return self._sys_db.list_application_versions()

    def get_latest_application_version(self) -> VersionInfo:
        """Return the latest application version."""
        return self._sys_db.get_latest_application_version()

    def set_latest_application_version(self, version_name: str) -> None:
        """Set a version as the latest by updating its timestamp to now."""
        new_timestamp = int(time.time() * 1000)
        self._sys_db.update_application_version_timestamp(version_name, new_timestamp)

    async def list_application_versions_async(self) -> List[VersionInfo]:
        """Async version of :meth:`list_application_versions`."""
        return await asyncio.to_thread(self.list_application_versions)

    async def get_latest_application_version_async(self) -> VersionInfo:
        """Async version of :meth:`get_latest_application_version`."""
        return await asyncio.to_thread(self.get_latest_application_version)

    async def set_latest_application_version_async(self, version_name: str) -> None:
        """Async version of :meth:`set_latest_application_version`."""
        await asyncio.to_thread(self.set_latest_application_version, version_name)



from typing import Optional

import sqlalchemy as sa
import typer

from dbos._app_db import ApplicationDatabase
from dbos._serialization import DefaultSerializer
from dbos._sys_db import SystemDatabase


def run_dbos_database_migrations(
    system_database_url: str,
    *,
    app_database_url: Optional[str] = None,
    schema: str = "dbos",
    application_role: Optional[str] = None,
) -> None:
    # First, run DBOS migrations on the system database and (optionally) the application database
    migrate_dbos_databases(
        system_database_url=system_database_url,
        app_database_url=app_database_url,
        schema=schema,
    )

    # Then, assign permissions on the DBOS schema to the application role, if any
    if application_role:
        if app_database_url:
            grant_dbos_schema_permissions(
                database_url=app_database_url,
                role_name=application_role,
                schema=schema,
            )
        grant_dbos_schema_permissions(
            database_url=system_database_url, role_name=application_role, schema=schema
        )


def migrate_dbos_databases(
    system_database_url: str, app_database_url: Optional[str], schema: str
) -> None:
    app_db = None
    sys_db = None
    try:
        sys_db = SystemDatabase.create(
            system_database_url=system_database_url,
            engine_kwargs={
                "pool_timeout": 30,
                "max_overflow": 0,
                "pool_size": 2,
            },
            engine=None,
            schema=schema,
            serializer=DefaultSerializer(),
            executor_id=None,
        )
        sys_db.run_migrations()
        if app_database_url:
            app_db = ApplicationDatabase.create(
                database_url=app_database_url,
                engine_kwargs={
                    "pool_timeout": 30,
                    "max_overflow": 0,
                    "pool_size": 2,
                },
                schema=schema,
                serializer=DefaultSerializer(),
            )
            app_db.run_migrations()
    except Exception as e:
        typer.echo(f"DBOS migrations failed: {e}")
        raise typer.Exit(code=1)
    finally:
        if sys_db:
            sys_db.destroy()
        if app_db:
            app_db.destroy()


def grant_dbos_schema_permissions(
    database_url: str, role_name: str, schema: str
) -> None:
    """
    Grant all permissions on all entities in the system schema to the specified role.
    """
    typer.echo(
        f"Granting permissions for the {schema} schema to {role_name} in database {sa.make_url(database_url)}"
    )
    engine = None
    try:
        engine = sa.create_engine(
            sa.make_url(database_url).set(drivername="postgresql+psycopg")
        )
        with engine.connect() as connection:
            connection.execution_options(isolation_level="AUTOCOMMIT")

            # Grant usage on the system schema
            sql = f'GRANT USAGE ON SCHEMA "{schema}" TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

            # Grant all privileges on all existing tables in the system schema (includes views)
            sql = f'GRANT ALL PRIVILEGES ON ALL TABLES IN SCHEMA "{schema}" TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

            # Grant all privileges on all sequences in the system schema
            sql = f'GRANT ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA "{schema}" TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

            # Grant execute on all functions and procedures in the system schema
            sql = (
                f'GRANT EXECUTE ON ALL FUNCTIONS IN SCHEMA "{schema}" TO "{role_name}"'
            )
            typer.echo(sql)
            connection.execute(sa.text(sql))

            # Grant default privileges for future objects in the system schema
            sql = f'ALTER DEFAULT PRIVILEGES IN SCHEMA "{schema}" GRANT ALL ON TABLES TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

            sql = f'ALTER DEFAULT PRIVILEGES IN SCHEMA "{schema}" GRANT ALL ON SEQUENCES TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

            sql = f'ALTER DEFAULT PRIVILEGES IN SCHEMA "{schema}" GRANT EXECUTE ON FUNCTIONS TO "{role_name}"'
            typer.echo(sql)
            connection.execute(sa.text(sql))

    except Exception as e:
        typer.echo(f"Failed to grant permissions to role {role_name}: {e}")
        raise typer.Exit(code=1)
    finally:
        if engine:
            engine.dispose()



import json
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Dict, List, Optional, Type, TypedDict, TypeVar, Union

from dbos._serialization import Serializer
from dbos._sys_db import StepInfo, VersionInfo, WorkflowSchedule, WorkflowStatus


class MessageType(str, Enum):
    EXECUTOR_INFO = "executor_info"
    RECOVERY = "recovery"
    CANCEL = "cancel"
    DELETE = "delete"
    LIST_WORKFLOWS = "list_workflows"
    LIST_QUEUED_WORKFLOWS = "list_queued_workflows"
    RESUME = "resume"
    RESTART = "restart"
    GET_WORKFLOW = "get_workflow"
    EXIST_PENDING_WORKFLOWS = "exist_pending_workflows"
    LIST_STEPS = "list_steps"
    FORK_WORKFLOW = "fork_workflow"
    RETENTION = "retention"
    GET_METRICS = "get_metrics"
    ALERT = "alert"
    EXPORT_WORKFLOW = "export_workflow"
    IMPORT_WORKFLOW = "import_workflow"
    LIST_SCHEDULES = "list_schedules"
    GET_SCHEDULE = "get_schedule"
    PAUSE_SCHEDULE = "pause_schedule"
    RESUME_SCHEDULE = "resume_schedule"
    BACKFILL_SCHEDULE = "backfill_schedule"
    TRIGGER_SCHEDULE = "trigger_schedule"
    LIST_APPLICATION_VERSIONS = "list_application_versions"
    SET_LATEST_APPLICATION_VERSION = "set_latest_application_version"


T = TypeVar("T", bound="BaseMessage")


@dataclass
class BaseMessage:
    type: MessageType
    request_id: str

    @classmethod
    def from_json(cls: Type[T], json_str: str) -> T:
        """
        Safely load a JSON into a dataclass, loading only the
        attributes specified in the dataclass.
        """
        data = json.loads(json_str)
        all_annotations = {}
        for base_cls in cls.__mro__:
            if hasattr(base_cls, "__annotations__"):
                all_annotations.update(base_cls.__annotations__)
        kwargs = {k: v for k, v in data.items() if k in all_annotations}
        return cls(**kwargs)

    def to_json(self) -> str:
        dict_data = asdict(self)
        return json.dumps(dict_data)


@dataclass
class BaseResponse(BaseMessage):
    error_message: Optional[str] = None


@dataclass
class ExecutorInfoRequest(BaseMessage):
    pass


@dataclass
class ExecutorInfoResponse(BaseMessage):
    executor_id: str
    application_version: str
    hostname: Optional[str]
    language: Optional[str]
    dbos_version: Optional[str]
    error_message: Optional[str] = None


@dataclass
class RecoveryRequest(BaseMessage):
    executor_ids: List[str]


@dataclass
class RecoveryResponse(BaseMessage):
    success: bool
    error_message: Optional[str] = None


@dataclass
class CancelRequest(BaseMessage):
    workflow_id: str


@dataclass
class CancelResponse(BaseMessage):
    success: bool
    error_message: Optional[str] = None


@dataclass
class DeleteRequest(BaseMessage):
    workflow_id: str
    delete_children: bool


@dataclass
class DeleteResponse(BaseMessage):
    success: bool
    error_message: Optional[str] = None


@dataclass
class ResumeRequest(BaseMessage):
    workflow_id: str


@dataclass
class ResumeResponse(BaseMessage):
    success: bool
    error_message: Optional[str] = None


@dataclass
class RestartRequest(BaseMessage):
    workflow_id: str


@dataclass
class RestartResponse(BaseMessage):
    success: bool
    error_message: Optional[str] = None


class ListWorkflowsBody(TypedDict, total=False):
    workflow_uuids: List[str]
    workflow_name: Optional[Union[str, List[str]]]
    authenticated_user: Optional[Union[str, List[str]]]
    start_time: Optional[str]
    end_time: Optional[str]
    status: Optional[Union[str, List[str]]]
    application_version: Optional[Union[str, List[str]]]
    forked_from: Optional[Union[str, List[str]]]
    parent_workflow_id: Optional[Union[str, List[str]]]
    queue_name: Optional[Union[str, List[str]]]
    limit: Optional[int]
    offset: Optional[int]
    sort_desc: bool
    workflow_id_prefix: Optional[Union[str, List[str]]]
    load_input: bool
    load_output: bool
    executor_id: Optional[Union[str, List[str]]]
    queues_only: bool


@dataclass
class WorkflowsOutput:
    WorkflowUUID: str
    Status: Optional[str]
    WorkflowName: Optional[str]
    WorkflowClassName: Optional[str]
    WorkflowConfigName: Optional[str]
    AuthenticatedUser: Optional[str]
    AssumedRole: Optional[str]
    AuthenticatedRoles: Optional[str]
    Input: Optional[str]
    Output: Optional[str]
    Error: Optional[str]
    CreatedAt: Optional[str]
    UpdatedAt: Optional[str]
    QueueName: Optional[str]
    ApplicationVersion: Optional[str]
    ExecutorID: Optional[str]
    WorkflowTimeoutMS: Optional[str]
    WorkflowDeadlineEpochMS: Optional[str]
    DeduplicationID: Optional[str]
    Priority: Optional[str]
    QueuePartitionKey: Optional[str]
    ForkedFrom: Optional[str]
    ParentWorkflowID: Optional[str]
    DequeuedAt: Optional[str]

    @classmethod
    def from_workflow_information(cls, info: WorkflowStatus) -> "WorkflowsOutput":
        # Convert fields to strings as needed
        created_at_str = str(info.created_at) if info.created_at is not None else None
        updated_at_str = str(info.updated_at) if info.updated_at is not None else None
        inputs_str = str(info.input) if info.input is not None else None
        outputs_str = str(info.output) if info.output is not None else None
        error_str = str(info.error) if info.error is not None else None
        roles_str = (
            str(info.authenticated_roles)
            if info.authenticated_roles is not None
            else None
        )
        workflow_timeout_ms_str = (
            str(info.workflow_timeout_ms)
            if info.workflow_timeout_ms is not None
            else None
        )
        workflow_deadline_epoch_ms_str = (
            str(info.workflow_deadline_epoch_ms)
            if info.workflow_deadline_epoch_ms is not None
            else None
        )
        priority_str = str(info.priority) if info.priority is not None else None
        dequeued_at_str = (
            str(info.dequeued_at) if info.dequeued_at is not None else None
        )

        return cls(
            WorkflowUUID=info.workflow_id,
            Status=info.status,
            WorkflowName=info.name,
            WorkflowClassName=info.class_name,
            WorkflowConfigName=info.config_name,
            AuthenticatedUser=info.authenticated_user,
            AssumedRole=info.assumed_role,
            AuthenticatedRoles=roles_str,
            Input=inputs_str,
            Output=outputs_str,
            Error=error_str,
            CreatedAt=created_at_str,
            UpdatedAt=updated_at_str,
            QueueName=info.queue_name,
            ApplicationVersion=info.app_version,
            ExecutorID=info.executor_id,
            WorkflowTimeoutMS=workflow_timeout_ms_str,
            WorkflowDeadlineEpochMS=workflow_deadline_epoch_ms_str,
            DeduplicationID=info.deduplication_id,
            Priority=priority_str,
            QueuePartitionKey=info.queue_partition_key,
            ForkedFrom=info.forked_from,
            ParentWorkflowID=info.parent_workflow_id,
            DequeuedAt=dequeued_at_str,
        )


@dataclass
class WorkflowSteps:
    function_id: int
    function_name: str
    output: Optional[str]
    error: Optional[str]
    child_workflow_id: Optional[str]
    started_at_epoch_ms: Optional[str]
    completed_at_epoch_ms: Optional[str]

    @classmethod
    def from_step_info(cls, info: StepInfo) -> "WorkflowSteps":
        output_str = str(info["output"]) if info["output"] is not None else None
        error_str = str(info["error"]) if info["error"] is not None else None
        started_at_str = (
            str(info["started_at_epoch_ms"])
            if info["started_at_epoch_ms"] is not None
            else None
        )
        completed_at_str = (
            str(info["completed_at_epoch_ms"])
            if info["completed_at_epoch_ms"] is not None
            else None
        )
        return cls(
            function_id=info["function_id"],
            function_name=info["function_name"],
            started_at_epoch_ms=started_at_str,
            completed_at_epoch_ms=completed_at_str,
            output=output_str,
            error=error_str,
            child_workflow_id=info["child_workflow_id"],
        )


@dataclass
class ListWorkflowsRequest(BaseMessage):
    body: ListWorkflowsBody


@dataclass
class ListWorkflowsResponse(BaseMessage):
    output: List[WorkflowsOutput]
    error_message: Optional[str] = None


class ListQueuedWorkflowsBody(TypedDict, total=False):
    workflow_uuids: List[str]
    workflow_name: Optional[Union[str, List[str]]]
    authenticated_user: Optional[Union[str, List[str]]]
    start_time: Optional[str]
    end_time: Optional[str]
    status: Optional[Union[str, List[str]]]
    application_version: Optional[Union[str, List[str]]]
    forked_from: Optional[Union[str, List[str]]]
    parent_workflow_id: Optional[Union[str, List[str]]]
    queue_name: Optional[Union[str, List[str]]]
    limit: Optional[int]
    offset: Optional[int]
    sort_desc: bool
    workflow_id_prefix: Optional[Union[str, List[str]]]
    load_input: bool
    load_output: bool
    executor_id: Optional[Union[str, List[str]]]


@dataclass
class ListQueuedWorkflowsRequest(BaseMessage):
    body: ListQueuedWorkflowsBody


@dataclass
class ListQueuedWorkflowsResponse(BaseMessage):
    output: List[WorkflowsOutput]
    error_message: Optional[str] = None


@dataclass
class GetWorkflowRequest(BaseMessage):
    workflow_id: str


@dataclass
class GetWorkflowResponse(BaseMessage):
    output: Optional[WorkflowsOutput]
    error_message: Optional[str] = None


@dataclass
class ExistPendingWorkflowsRequest(BaseMessage):
    executor_id: str
    application_version: str


@dataclass
class ExistPendingWorkflowsResponse(BaseMessage):
    exist: bool
    error_message: Optional[str] = None


@dataclass
class ListStepsRequest(BaseMessage):
    workflow_id: str


@dataclass
class ListStepsResponse(BaseMessage):
    output: Optional[List[WorkflowSteps]]
    error_message: Optional[str] = None


class ForkWorkflowBody(TypedDict):
    workflow_id: str
    start_step: int
    application_version: Optional[str]
    new_workflow_id: Optional[str]


@dataclass
class ForkWorkflowRequest(BaseMessage):
    body: ForkWorkflowBody


@dataclass
class ForkWorkflowResponse(BaseMessage):
    new_workflow_id: Optional[str]
    error_message: Optional[str] = None


class RetentionBody(TypedDict):
    gc_cutoff_epoch_ms: Optional[int]
    gc_rows_threshold: Optional[int]
    timeout_cutoff_epoch_ms: Optional[int]


@dataclass
class RetentionRequest(BaseMessage):
    body: RetentionBody


@dataclass
class RetentionResponse(BaseMessage):
    success: bool
    error_message: Optional[str] = None


@dataclass
class GetMetricsRequest(BaseMessage):
    start_time: str  # ISO 8601
    end_time: str  # ISO 8601
    metric_class: str


@dataclass
class MetricData:
    metric_type: str
    metric_name: str
    value: int


@dataclass
class GetMetricsResponse(BaseMessage):
    metrics: List[MetricData]
    error_message: Optional[str] = None


@dataclass
class AlertRequest(BaseMessage):
    name: str
    message: str
    metadata: Dict[str, str]


@dataclass
class AlertResponse(BaseMessage):
    success: bool
    error_message: Optional[str] = None


@dataclass
class ExportWorkflowRequest(BaseMessage):
    workflow_id: str
    export_children: bool


@dataclass
class ExportWorkflowResponse(BaseMessage):
    serialized_workflow: Optional[str]
    error_message: Optional[str] = None


@dataclass
class ImportWorkflowRequest(BaseMessage):
    serialized_workflow: str


@dataclass
class ImportWorkflowResponse(BaseMessage):
    success: bool
    error_message: Optional[str] = None


# --- Schedule messages ---


@dataclass
class ScheduleOutput:
    schedule_id: str
    schedule_name: str
    workflow_name: str
    workflow_class_name: Optional[str]
    schedule: str
    status: str
    context: str

    @classmethod
    def from_schedule(
        cls, s: WorkflowSchedule, serializer: Serializer
    ) -> "ScheduleOutput":
        context_str = str(serializer.deserialize(s["context"]))
        return cls(
            schedule_id=s["schedule_id"],
            schedule_name=s["schedule_name"],
            workflow_name=s["workflow_name"],
            workflow_class_name=s["workflow_class_name"],
            schedule=s["schedule"],
            status=s["status"],
            context=context_str,
        )


class ListSchedulesBody(TypedDict, total=False):
    status: Optional[Union[str, List[str]]]
    workflow_name: Optional[Union[str, List[str]]]
    schedule_name_prefix: Optional[Union[str, List[str]]]


@dataclass
class ListSchedulesRequest(BaseMessage):
    body: ListSchedulesBody


@dataclass
class ListSchedulesResponse(BaseMessage):
    output: List[ScheduleOutput]
    error_message: Optional[str] = None


@dataclass
class GetScheduleRequest(BaseMessage):
    schedule_name: str


@dataclass
class GetScheduleResponse(BaseMessage):
    output: Optional[ScheduleOutput]
    error_message: Optional[str] = None


@dataclass
class PauseScheduleRequest(BaseMessage):
    schedule_name: str


@dataclass
class PauseScheduleResponse(BaseMessage):
    success: bool
    error_message: Optional[str] = None


@dataclass
class ResumeScheduleRequest(BaseMessage):
    schedule_name: str


@dataclass
class ResumeScheduleResponse(BaseMessage):
    success: bool
    error_message: Optional[str] = None


@dataclass
class BackfillScheduleRequest(BaseMessage):
    schedule_name: str
    start: str  # ISO 8601
    end: str  # ISO 8601


@dataclass
class BackfillScheduleResponse(BaseMessage):
    workflow_ids: List[str]
    error_message: Optional[str] = None


@dataclass
class TriggerScheduleRequest(BaseMessage):
    schedule_name: str


@dataclass
class TriggerScheduleResponse(BaseMessage):
    workflow_id: Optional[str]
    error_message: Optional[str] = None


# --- Application version messages ---


@dataclass
class ApplicationVersionOutput:
    version_id: str
    version_name: str
    version_timestamp: int
    created_at: int

    @classmethod
    def from_version_info(cls, v: VersionInfo) -> "ApplicationVersionOutput":
        return cls(
            version_id=v["version_id"],
            version_name=v["version_name"],
            version_timestamp=v["version_timestamp"],
            created_at=v["created_at"],
        )


@dataclass
class ListApplicationVersionsRequest(BaseMessage):
    pass


@dataclass
class ListApplicationVersionsResponse(BaseMessage):
    output: List[ApplicationVersionOutput]
    error_message: Optional[str] = None


@dataclass
class SetLatestApplicationVersionRequest(BaseMessage):
    version_name: str


@dataclass
class SetLatestApplicationVersionResponse(BaseMessage):
    success: bool
    error_message: Optional[str] = None



import random
import threading
import traceback
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Callable, Coroutine

from ._logger import dbos_logger

if TYPE_CHECKING:
    from ._dbos import DBOSRegistry

from ._context import SetWorkflowID
from ._croniter import croniter  # type: ignore
from ._registrations import get_dbos_func_name

DecoratedScheduledWorkflow = (
    Callable[[datetime, datetime], None]
    | Callable[[datetime, datetime], Coroutine[Any, Any, None]]
)


def decorator_scheduler_loop(
    func: DecoratedScheduledWorkflow, cron: str, stop_event: threading.Event
) -> None:
    from dbos._dbos import _get_dbos_instance

    dbos = _get_dbos_instance()
    scheduler_queue = dbos._registry.get_internal_queue()
    try:
        iter = croniter(cron, datetime.now(timezone.utc), second_at_beginning=True)
    except Exception:
        dbos_logger.error(
            f'Cannot run scheduled function {get_dbos_func_name(func)}. Invalid crontab "{cron}"'
        )
        raise
    while not stop_event.is_set():
        next_exec_time = iter.get_next(datetime)
        sleep_time = (next_exec_time - datetime.now(timezone.utc)).total_seconds()
        sleep_time = max(0, sleep_time)
        # To prevent a "thundering herd" problem in a distributed setting,
        # apply jitter of up to 10% the sleep time, capped at 10 seconds
        max_jitter = min(sleep_time / 10, 10)
        jitter = random.uniform(0, max_jitter)
        if stop_event.wait(timeout=sleep_time + jitter):
            return
        try:
            workflowID = (
                f"sched-{get_dbos_func_name(func)}-{next_exec_time.isoformat()}"
            )
            if not dbos._sys_db.get_workflow_status(workflowID):
                with SetWorkflowID(workflowID):
                    scheduler_queue.enqueue(
                        func, next_exec_time, datetime.now(timezone.utc)
                    )
        except Exception:
            dbos_logger.warning(
                f"Exception encountered in scheduler thread: {traceback.format_exc()})"
            )


def scheduled(
    dbosreg: "DBOSRegistry", cron: str
) -> Callable[[DecoratedScheduledWorkflow], DecoratedScheduledWorkflow]:
    def decorator(func: DecoratedScheduledWorkflow) -> DecoratedScheduledWorkflow:
        try:
            croniter(cron, datetime.now(timezone.utc), second_at_beginning=True)
        except Exception:
            raise ValueError(
                f'Invalid crontab "{cron}" for scheduled function function {get_dbos_func_name(func)}.'
            )
        stop_event = threading.Event()
        dbosreg.register_poller(
            stop_event, decorator_scheduler_loop, func, cron, stop_event
        )
        return func

    return decorator



import asyncio
import threading
import time
import uuid
from concurrent.futures import Future, ThreadPoolExecutor
from dataclasses import dataclass
from typing import Any, Callable, ClassVar, Coroutine, List, Tuple, Union, cast

import pytest
from sqlalchemy import text

# Public API
from dbos import DBOS, SetWorkflowID, StepInfo
from dbos._dbos_config import DBOSConfig
from dbos._error import DBOSNonExistentWorkflowError
from tests.conftest import using_sqlite


def test_concurrent_workflows(dbos: DBOS) -> None:
    @DBOS.workflow()
    def test_workflow() -> str:
        time.sleep(1)
        workflow_id = DBOS.workflow_id
        assert workflow_id is not None
        return workflow_id

    def test_thread(id: str) -> str:
        with SetWorkflowID(id):
            return test_workflow()

    num_threads = 10
    with ThreadPoolExecutor(max_workers=num_threads) as executor:
        futures: list[Tuple[str, Future[str]]] = []
        for _ in range(num_threads):
            id = str(uuid.uuid4())
            futures.append((id, executor.submit(test_thread, id)))
        for id, future in futures:
            assert id == future.result()


def test_concurrent_getevent(dbos: DBOS) -> None:
    @DBOS.workflow()
    def test_workflow(event_name: str, value: str) -> str:
        DBOS.set_event(event_name, value)
        return value

    def test_thread(id: str, event_name: str) -> str:
        return cast(str, DBOS.get_event(id, event_name, 5))

    wfuuid = str(uuid.uuid4())
    event_name = "test_event"
    with ThreadPoolExecutor(max_workers=2) as executor:
        future1 = executor.submit(test_thread, wfuuid, event_name)
        future2 = executor.submit(test_thread, wfuuid, event_name)

        expected_message = "test message"
        with SetWorkflowID(wfuuid):
            test_workflow(event_name, expected_message)

        # Both should return the same message (shared handle, NOT concurrent)
        assert future1.result() == future2.result()
        assert future1.result() == expected_message
        # Make sure the event map is empty
        assert not dbos._sys_db.workflow_events_map._dict


# Async concurrency tests


@dataclass(frozen=True)
class Thing:
    func: Callable[[], Coroutine[Any, Any, str]]
    expected: str


async def run_things_serial_or_conc(conc: bool, things: List[Thing]) -> None:
    if conc:
        # `asyncio.gather` without return_exceptions is harmful (like TS Promise.all or race)
        #    return_exceptions, which will wait for all, capture exceptions as values, and is OK.
        tasks: List[asyncio.Task[str]] = [asyncio.create_task(t.func()) for t in things]

        results: List[Union[str, BaseException]] = await asyncio.gather(
            *tasks,
            return_exceptions=True,
        )

        for i, (thing, result) in enumerate(zip(things, results)):
            if isinstance(result, BaseException):
                # Make failures loud
                raise AssertionError(
                    f"Thing[{i}] raised {type(result).__name__}: {result}"
                ) from result
            assert (
                result == thing.expected
            ), f"Thing[{i}] expected {thing.expected!r}, got {result!r}"

    else:
        for i, thing in enumerate(things):
            result = await thing.func()
            assert (
                result == thing.expected
            ), f"Thing[{i}] expected {thing.expected!r}, got {result!r}"


def compare_wf_runs(
    wfsteps_serial: List[StepInfo], wfsteps_concurrent: List[StepInfo]
) -> None:
    iconc = 0
    i = 0
    while i < len(wfsteps_serial):
        s = wfsteps_serial[i]
        c = wfsteps_concurrent[iconc]

        print(
            f"Output of {c['function_name']}@{c['function_id']}: {c['output']!r} "
            f"vs. {s['function_name']}@{s['function_id']}: {s['output']!r}"
        )

        # Allow extra DBOS.sleep in concurrent run that may interleave
        if c["function_id"] < s["function_id"] and c["function_name"] == "DBOS.sleep":
            iconc += 1
            continue

        assert c["function_id"] == s["function_id"]
        assert c["function_name"] == s["function_name"]
        assert c["error"] == s["error"]

        if c["function_name"] in {"DBOS.now", "DBOS.randomUUID", "DBOS.sleep"}:
            # output may differ
            pass
        else:
            assert c["output"] == s["output"]

        i += 1
        iconc += 1


@pytest.mark.asyncio
async def test_gather_manythings(dbos: DBOS) -> None:
    @DBOS.workflow()
    async def simple_wf() -> str:
        return "WF Ran"

    @DBOS.step()
    async def simple_step() -> str:
        return "Step Ran"

    @DBOS.dbos_class()
    class ConcurrTestClass:
        cnt: ClassVar[int] = 0

        @staticmethod
        @DBOS.step()
        async def test_read_write_function(id: int) -> int:
            ConcurrTestClass.cnt += 1
            return id

        @staticmethod
        @DBOS.step()
        async def test_step(id: int) -> int:
            ConcurrTestClass.cnt += 1
            return id

        @staticmethod
        @DBOS.step()
        async def test_step_str(id: str) -> str:
            return id

        @staticmethod
        @DBOS.step(
            retries_allowed=True, max_attempts=5, interval_seconds=0.01, backoff_rate=1
        )
        async def test_step_retry(id: str) -> str:
            ss = DBOS.step_status
            assert ss is not None
            assert ss.current_attempt is not None
            if ss.current_attempt <= 2:
                raise Exception("Not yet")
            return id

        @staticmethod
        @DBOS.workflow()
        async def receive_workflow(topic: str, timeout: int) -> Any:
            return await DBOS.recv_async(topic, timeout)

    @DBOS.workflow()
    async def run_a_lot_of_things_at_once(conc: bool) -> None:
        async def t_sleep() -> str:
            await DBOS.sleep_async(0.5)
            return "slept"

        async def t_run_step1() -> str:
            return await DBOS.run_step_async({"name": "runStep1"}, lambda: "ranStep")

        async def t_run_step2() -> str:
            return await DBOS.run_step_async({"name": "runStep2"}, lambda: "ranStep")

        async def t_run_step_retry() -> str:
            def body() -> str:
                ss = DBOS.step_status
                assert ss is not None
                assert ss.current_attempt is not None
                if ss.current_attempt <= 2:
                    raise Exception("Not yet")
                return "ranStep"

            return await DBOS.run_step_async(
                {
                    "name": "runStepRetry",
                    "retries_allowed": True,
                    "max_attempts": 5,
                    "interval_seconds": 0.01,
                    "backoff_rate": 1,
                },
                body,
            )

        async def t_function_calls_function() -> str:
            return str(await ConcurrTestClass.test_read_write_function(2))

        async def t_set_event() -> str:
            await DBOS.set_event_async("eventkey", "eval")
            return "set"

        async def t_get_event() -> str:
            wfid = DBOS.workflow_id
            assert wfid is not None
            v = await DBOS.get_event_async(wfid, "eventkey")
            return v if v is not None else "Nope"

        async def t_send_msg() -> str:
            wfid = DBOS.workflow_id
            assert wfid is not None
            await DBOS.send_async(wfid, "msg", "topic")
            return "sent"

        async def t_step_str_3() -> str:
            return await ConcurrTestClass.test_step_str("3")

        async def t_recv_msg() -> str:
            v = await DBOS.recv_async("topic")
            return v if v is not None else "Nope"

        async def t_get_workflow_status_nosuch() -> str:
            st = await DBOS.get_workflow_status_async("nosuchwv")
            return st.status if st is not None else "Nope"

        async def t_retrieve_workflow_nosuch() -> str:
            with pytest.raises(DBOSNonExistentWorkflowError):
                await DBOS.retrieve_workflow_async("nosuchwv", existing_workflow=True)
                return "Yep"
            return "Nope"

        async def t_liststeps() -> str:
            steps = await DBOS.list_workflow_steps_async("nosuchwv")
            return f"{len(steps)}"

        async def t_listwfs() -> str:
            wfs = await DBOS.list_workflows_async(workflow_id_prefix="aaaa")
            return f"{len(wfs)}"

        async def t_listqwfs() -> str:
            wfs = await DBOS.list_queued_workflows_async(workflow_id_prefix="aaaa")
            return f"{len(wfs)}"

        async def t_step_retry_4() -> str:
            return await ConcurrTestClass.test_step_retry("4")

        async def t_start_child() -> str:
            with SetWorkflowID(f"{DBOS.workflow_id}-cwf"):
                await DBOS.start_workflow_async(simple_wf)
            return "started"

        async def t_get_child_result() -> str:
            res = await DBOS.get_result_async(f"{DBOS.workflow_id}-cwf")
            assert res is not None
            return cast(str, res)

        async def t_write_stream() -> str:
            await DBOS.write_stream_async("stream", "val")
            return "wrote"

        async def t_read_stream() -> str:
            wfid = DBOS.workflow_id
            assert wfid is not None
            ait = DBOS.read_stream_async(wfid, "stream")
            item = await ait.__anext__()
            return cast(str, item)

        things: List[Thing] = [
            Thing(func=t_sleep, expected="slept"),
            Thing(func=t_run_step1, expected="ranStep"),
            Thing(func=t_run_step2, expected="ranStep"),
            Thing(func=t_listqwfs, expected="0"),
            Thing(func=t_run_step_retry, expected="ranStep"),
            Thing(func=t_function_calls_function, expected="2"),
            Thing(func=t_set_event, expected="set"),
            Thing(func=t_get_event, expected="eval"),
            Thing(func=t_send_msg, expected="sent"),
            Thing(func=t_step_str_3, expected="3"),
            Thing(func=t_recv_msg, expected="msg"),
            Thing(func=t_get_workflow_status_nosuch, expected="Nope"),
            Thing(func=t_listwfs, expected="0"),
            Thing(func=t_retrieve_workflow_nosuch, expected="Nope"),
            Thing(func=t_step_retry_4, expected="4"),
            Thing(func=simple_wf, expected="WF Ran"),
            Thing(func=t_start_child, expected="started"),
            Thing(func=t_get_child_result, expected="WF Ran"),
            Thing(func=simple_step, expected="Step Ran"),
            Thing(func=t_write_stream, expected="wrote"),
            Thing(func=t_read_stream, expected="val"),
            Thing(func=t_liststeps, expected="0"),
        ]

        await run_things_serial_or_conc(conc, things)

    wfid_serial = str(uuid.uuid4())
    wfid_concurrent = str(uuid.uuid4())

    with SetWorkflowID(wfid_serial):
        await run_a_lot_of_things_at_once(conc=False)  # from earlier translation

    with SetWorkflowID(wfid_concurrent):
        await run_a_lot_of_things_at_once(conc=True)

    wfsteps_serial = await DBOS.list_workflow_steps_async(wfid_serial)
    wfsteps_concurrent = await DBOS.list_workflow_steps_async(wfid_concurrent)

    assert wfsteps_serial is not None
    assert wfsteps_concurrent is not None
    compare_wf_runs(wfsteps_serial, wfsteps_concurrent)

    await run_a_lot_of_things_at_once(conc=True)


@pytest.mark.asyncio
async def test_gather_manysteps(dbos: DBOS) -> None:
    @DBOS.dbos_class()
    class ConcurrTestClass:
        @staticmethod
        @DBOS.step()
        async def test_step_str(id: str) -> str:
            return id

        @staticmethod
        @DBOS.step(
            retries_allowed=True, max_attempts=5, interval_seconds=0.01, backoff_rate=1
        )
        async def test_step_retry(id: str) -> str:
            ss = DBOS.step_status
            assert ss is not None
            assert ss.current_attempt is not None
            if ss.current_attempt <= 2:
                raise Exception("Not yet")
            return id

    @DBOS.workflow()
    async def run_a_lot_of_steps_at_once(conc: bool) -> None:
        async def t_run_step1a() -> str:
            return await DBOS.run_step_async({"name": "runStep1a"}, lambda: "ranStep")

        async def _retry_body() -> str:
            ss = DBOS.step_status
            assert ss is not None
            assert ss.current_attempt is not None
            if ss.current_attempt <= 2:
                raise Exception("Not yet")
            return "ranStep"

        async def t_run_step_retry1() -> str:
            return await DBOS.run_step_async(
                {
                    "name": "runStepRetry1",
                    "retries_allowed": True,
                    "max_attempts": 5,
                    "interval_seconds": 0.02,
                    "backoff_rate": 1,
                },
                _retry_body,
            )

        async def t_run_step_retry2() -> str:
            return await DBOS.run_step_async(
                {
                    "name": "runStepRetry2",
                    "retries_allowed": True,
                    "max_attempts": 5,
                    "interval_seconds": 0.01,
                    "backoff_rate": 1,
                },
                _retry_body,
            )

        async def t_run_step2() -> str:
            return await DBOS.run_step_async({"name": "runStep2"}, lambda: "ranStep")

        async def t_step_str_3() -> str:
            return await ConcurrTestClass.test_step_str("3")

        async def t_run_step_retry3() -> str:
            return await DBOS.run_step_async(
                {
                    "name": "runStepRetry3",
                    "retries_allowed": True,
                    "max_attempts": 5,
                    "interval_seconds": 0.03,
                    "backoff_rate": 1,
                },
                _retry_body,
            )

        async def t_step_retry_4() -> str:
            return await ConcurrTestClass.test_step_retry("4")

        things: List[Thing] = [
            Thing(func=t_run_step1a, expected="ranStep"),
            Thing(func=t_run_step_retry1, expected="ranStep"),
            Thing(func=t_run_step_retry2, expected="ranStep"),
            Thing(func=t_run_step2, expected="ranStep"),
            Thing(func=t_step_str_3, expected="3"),
            Thing(func=t_run_step_retry3, expected="ranStep"),
            Thing(func=t_step_retry_4, expected="4"),
            # TODO: Add some sync steps?
        ]

        await run_things_serial_or_conc(conc, things)

    wfid_serial = str(uuid.uuid4())
    wfid_concurrent = str(uuid.uuid4())

    with SetWorkflowID(wfid_serial):
        await run_a_lot_of_steps_at_once(conc=False)

    with SetWorkflowID(wfid_concurrent):
        await run_a_lot_of_steps_at_once(conc=True)

    wfsteps_serial = await DBOS.list_workflow_steps_async(wfid_serial)
    wfsteps_concurrent = await DBOS.list_workflow_steps_async(wfid_concurrent)

    assert wfsteps_serial is not None
    assert wfsteps_concurrent is not None
    compare_wf_runs(wfsteps_serial, wfsteps_concurrent)

    fwf1 = await DBOS.fork_workflow_async(wfid_concurrent, 3)
    await fwf1.get_result()

    fwf2 = await DBOS.fork_workflow_async(wfid_concurrent, 5)
    await fwf2.get_result()

    fwf3 = await DBOS.fork_workflow_async(wfid_concurrent, 7)
    await fwf3.get_result()


@pytest.mark.asyncio
async def test_gather_many_send_async(dbos: DBOS) -> None:
    num_messages = 100

    @DBOS.workflow()
    async def run_many_sends(conc: bool) -> None:
        things: List[Thing] = []

        for i in range(num_messages):
            topic = f"topic_{i}"
            msg = f"msg_{i}"

            async def t_send(t: str = topic, m: str = msg) -> str:
                wfid = DBOS.workflow_id
                assert wfid is not None
                await DBOS.send_async(wfid, m, t)
                return "sent"

            things.append(Thing(func=t_send, expected="sent"))

        await run_things_serial_or_conc(conc, things)

        for i in range(num_messages):
            v = await DBOS.recv_async(f"topic_{i}")
            assert v == f"msg_{i}"

    wfid_serial = str(uuid.uuid4())
    wfid_concurrent = str(uuid.uuid4())

    with SetWorkflowID(wfid_serial):
        await run_many_sends(conc=False)

    with SetWorkflowID(wfid_concurrent):
        await run_many_sends(conc=True)

    wfsteps_serial = await DBOS.list_workflow_steps_async(wfid_serial)
    wfsteps_concurrent = await DBOS.list_workflow_steps_async(wfid_concurrent)

    assert wfsteps_serial is not None
    assert wfsteps_concurrent is not None
    compare_wf_runs(wfsteps_serial, wfsteps_concurrent)


@pytest.mark.asyncio
async def test_gather_many_write_stream(dbos: DBOS) -> None:
    num_writes = 100

    @DBOS.workflow()
    async def run_many_writes(conc: bool) -> None:
        things: List[Thing] = []

        for i in range(num_writes):
            val = f"val_{i}"

            async def t_write(v: str = val) -> str:
                await DBOS.write_stream_async("stream", v)
                return "wrote"

            things.append(Thing(func=t_write, expected="wrote"))

        await run_things_serial_or_conc(conc, things)

    wfid_serial = str(uuid.uuid4())
    wfid_concurrent = str(uuid.uuid4())

    with SetWorkflowID(wfid_serial):
        await run_many_writes(conc=False)

    values_serial: List[str] = []
    async for item in DBOS.read_stream_async(wfid_serial, "stream"):
        values_serial.append(cast(str, item))
    assert set(values_serial) == {f"val_{i}" for i in range(num_writes)}

    with SetWorkflowID(wfid_concurrent):
        await run_many_writes(conc=True)

    values_concurrent: List[str] = []
    async for item in DBOS.read_stream_async(wfid_concurrent, "stream"):
        values_concurrent.append(cast(str, item))
    assert set(values_concurrent) == {f"val_{i}" for i in range(num_writes)}

    wfsteps_serial = await DBOS.list_workflow_steps_async(wfid_serial)
    wfsteps_concurrent = await DBOS.list_workflow_steps_async(wfid_concurrent)

    assert wfsteps_serial is not None
    assert wfsteps_concurrent is not None
    compare_wf_runs(wfsteps_serial, wfsteps_concurrent)


@pytest.mark.asyncio
async def test_gather_many_write_stream_from_step(dbos: DBOS) -> None:
    num_writes = 100

    @DBOS.step()
    async def write_stream_step(val: str) -> str:
        await DBOS.write_stream_async("stream", val)
        return "wrote"

    @DBOS.workflow()
    async def run_many_writes(conc: bool) -> None:
        things: List[Thing] = []

        for i in range(num_writes):
            val = f"val_{i}"

            async def t_write(v: str = val) -> str:
                return await write_stream_step(v)

            things.append(Thing(func=t_write, expected="wrote"))

        await run_things_serial_or_conc(conc, things)

    wfid_serial = str(uuid.uuid4())
    wfid_concurrent = str(uuid.uuid4())

    with SetWorkflowID(wfid_serial):
        await run_many_writes(conc=False)

    values_serial: List[str] = []
    async for item in DBOS.read_stream_async(wfid_serial, "stream"):
        values_serial.append(cast(str, item))
    assert set(values_serial) == {f"val_{i}" for i in range(num_writes)}

    with SetWorkflowID(wfid_concurrent):
        await run_many_writes(conc=True)

    values_concurrent: List[str] = []
    async for item in DBOS.read_stream_async(wfid_concurrent, "stream"):
        values_concurrent.append(cast(str, item))
    assert set(values_concurrent) == {f"val_{i}" for i in range(num_writes)}

    wfsteps_serial = await DBOS.list_workflow_steps_async(wfid_serial)
    wfsteps_concurrent = await DBOS.list_workflow_steps_async(wfid_concurrent)

    assert wfsteps_serial is not None
    assert wfsteps_concurrent is not None
    compare_wf_runs(wfsteps_serial, wfsteps_concurrent)


@pytest.mark.asyncio
async def test_gather_many_set_event(dbos: DBOS) -> None:
    num_events = 100

    @DBOS.workflow()
    async def run_many_set_events(conc: bool) -> None:
        things: List[Thing] = []

        for i in range(num_events):
            key = f"event_{i}"
            val = f"val_{i}"

            async def t_set(k: str = key, v: str = val) -> str:
                await DBOS.set_event_async(k, v)
                return "set"

            things.append(Thing(func=t_set, expected="set"))

        await run_things_serial_or_conc(conc, things)

    wfid_serial = str(uuid.uuid4())
    wfid_concurrent = str(uuid.uuid4())

    with SetWorkflowID(wfid_serial):
        await run_many_set_events(conc=False)

    for i in range(num_events):
        v = await DBOS.get_event_async(wfid_serial, f"event_{i}")
        assert v == f"val_{i}"

    with SetWorkflowID(wfid_concurrent):
        await run_many_set_events(conc=True)

    for i in range(num_events):
        v = await DBOS.get_event_async(wfid_concurrent, f"event_{i}")
        assert v == f"val_{i}"

    wfsteps_serial = await DBOS.list_workflow_steps_async(wfid_serial)
    wfsteps_concurrent = await DBOS.list_workflow_steps_async(wfid_concurrent)

    assert wfsteps_serial is not None
    assert wfsteps_concurrent is not None
    compare_wf_runs(wfsteps_serial, wfsteps_concurrent)


@pytest.mark.asyncio
async def test_high_async_concurrency(dbos: DBOS, config: DBOSConfig) -> None:
    config["max_executor_threads"] = 32
    DBOS.destroy(destroy_registry=True)
    DBOS(config=config)
    DBOS.launch()

    peak_threads = 0

    @DBOS.step()
    async def sleep_step() -> None:
        nonlocal peak_threads
        peak_threads = max(peak_threads, threading.active_count())
        await asyncio.sleep(5)

    @DBOS.workflow()
    async def sleep_workflow() -> None:
        assert DBOS.workflow_id
        nonlocal peak_threads
        peak_threads = max(peak_threads, threading.active_count())
        await DBOS.sleep_async(5)
        message = await DBOS.recv_async(timeout_seconds=5)
        assert message is None
        value = await DBOS.get_event_async(DBOS.workflow_id, "key", timeout_seconds=5)
        assert value is None

    @DBOS.workflow()
    async def concurrent_step_workflow() -> None:
        await asyncio.gather(*[sleep_step() for _ in range(1000)])

    @DBOS.workflow()
    async def concurrent_wf_workflow() -> None:
        handles = await asyncio.gather(
            *[DBOS.start_workflow_async(sleep_workflow) for _ in range(1000)]
        )
        await asyncio.gather(*[h.get_result() for h in handles])

    # Test concurrent steps
    start_time = time.time()
    await concurrent_step_workflow()
    elapsed = time.time() - start_time
    print(f"Steps elapsed: {elapsed}, peak threads: {peak_threads}")
    assert (
        config["max_executor_threads"]
        and peak_threads < config["max_executor_threads"] + 10
    )
    assert elapsed < 30

    # Test concurrent workflows
    start_time = time.time()
    await concurrent_wf_workflow()
    elapsed = time.time() - start_time
    print(f"Workflows elapsed: {elapsed}, peak threads: {peak_threads}")
    assert (
        config["max_executor_threads"]
        and peak_threads < config["max_executor_threads"] + 10
    )
    assert elapsed < 60



import asyncio
import copy
import functools
import inspect
import json
import sys
import threading
import time
import uuid
from concurrent.futures import Future
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Generic,
    List,
    Optional,
    ParamSpec,
    TypedDict,
    TypeVar,
    Union,
    cast,
)

from dbos._outcome import NoResult, Outcome, Pending
from dbos._utils import GlobalParams, retriable_postgres_exception

from ._app_db import ApplicationDatabase, TransactionResultInternal
from ._context import (
    DBOSAssumeRole,
    DBOSContext,
    DBOSContextEnsure,
    DBOSContextSetAuth,
    EnterDBOSStepCtx,
    EnterDBOSTransaction,
    EnterDBOSWorkflow,
    OperationType,
    SetWorkflowID,
    TracedAttributes,
    assert_current_dbos_context,
    get_local_dbos_context,
)
from ._error import (
    DBOSAwaitedWorkflowCancelledError,
    DBOSException,
    DBOSMaxStepRetriesExceeded,
    DBOSNonExistentWorkflowError,
    DBOSQueueDeduplicatedError,
    DBOSRecoveryError,
    DBOSUnexpectedStepError,
    DBOSWorkflowCancelledError,
    DBOSWorkflowConflictIDError,
    DBOSWorkflowFunctionNotFoundError,
)
from ._logger import dbos_logger
from ._registrations import (
    DEFAULT_MAX_RECOVERY_ATTEMPTS,
    DBOSFuncType,
    ValidateArgsCallable,
    get_config_name,
    get_dbos_class_name,
    get_dbos_func_name,
    get_func_info,
    get_or_create_func_info,
    set_dbos_func_name,
    set_func_info,
    set_temp_workflow_type,
)
from ._roles import check_required_roles
from ._serialization import (
    DBOSPortableJSON,
    WorkflowInputs,
    WorkflowSerializationFormat,
    deserialize_args,
    deserialize_exception,
    deserialize_value,
    serialize_args,
    serialize_exception,
    serialize_value,
    serialize_value_as,
)
from ._sys_db import (
    EnqueueOptionsInternal,
    OperationResultInternal,
    WorkflowStatus,
    WorkflowStatusInternal,
    WorkflowStatusString,
)

if TYPE_CHECKING:
    from ._dbos import (
        DBOS,
        WorkflowHandle,
        WorkflowHandleAsync,
        DBOSRegistry,
        IsolationLevel,
    )

from sqlalchemy.exc import DBAPIError, InvalidRequestError

P = ParamSpec("P")  # A generic type for workflow parameters
R = TypeVar("R", covariant=True)  # A generic type for workflow return values
F = TypeVar("F", bound=Callable[..., Any])

TEMP_SEND_WF_NAME = "<temp>.temp_send_workflow"
DEBOUNCER_WORKFLOW_NAME = "_dbos_debouncer_workflow"
DEFAULT_POLLING_INTERVAL = 1.0


class WorkflowHandleFuture(Generic[R]):

    def __init__(self, workflow_id: str, future: Future[R], dbos: "DBOS"):
        self.workflow_id = workflow_id
        self.future = future
        self.dbos = dbos

    def get_workflow_id(self) -> str:
        return self.workflow_id

    def get_result(
        self, *, polling_interval_sec: float = DEFAULT_POLLING_INTERVAL
    ) -> R:
        try:
            r = self.future.result()
        except Exception as e:
            serialized_e, serialization = serialize_exception(
                e, None, self.dbos._serializer
            )
            self.dbos._sys_db.record_get_result(
                self.workflow_id, None, serialized_e, serialization
            )
            raise
        serialized_r, serialization = serialize_value(r, None, self.dbos._serializer)
        self.dbos._sys_db.record_get_result(
            self.workflow_id, serialized_r, None, serialization
        )
        return r

    def get_status(self) -> WorkflowStatus:
        stat = self.dbos.get_workflow_status(self.workflow_id)
        if stat is None:
            raise DBOSNonExistentWorkflowError("target", self.workflow_id)
        return stat


class WorkflowHandlePolling(Generic[R]):

    def __init__(self, workflow_id: str, dbos: "DBOS"):
        self.workflow_id = workflow_id
        self.dbos = dbos

    def get_workflow_id(self) -> str:
        return self.workflow_id

    def get_result(
        self, *, polling_interval_sec: float = DEFAULT_POLLING_INTERVAL
    ) -> R:
        try:
            r: R = self.dbos._sys_db.await_workflow_result(
                self.workflow_id, polling_interval_sec
            )
        except Exception as e:
            serialized_e, serialization = serialize_exception(
                e, None, self.dbos._serializer
            )
            self.dbos._sys_db.record_get_result(
                self.workflow_id, None, serialized_e, serialization
            )
            raise
        serialized_r, serialization = serialize_value(r, None, self.dbos._serializer)
        self.dbos._sys_db.record_get_result(
            self.workflow_id, serialized_r, None, serialization
        )
        return r

    def get_status(self) -> WorkflowStatus:
        stat = self.dbos.get_workflow_status(self.workflow_id)
        if stat is None:
            raise DBOSNonExistentWorkflowError("target", self.workflow_id)
        return stat


class WorkflowHandleAsyncTask(Generic[R]):

    def __init__(self, workflow_id: str, task: asyncio.Future[R], dbos: "DBOS"):
        self.workflow_id = workflow_id
        self.task = task
        self.dbos = dbos

    def get_workflow_id(self) -> str:
        return self.workflow_id

    async def get_result(
        self, *, polling_interval_sec: float = DEFAULT_POLLING_INTERVAL
    ) -> R:
        try:
            r = await self.task
        except Exception as e:
            serialized_e, serialization = serialize_exception(
                e, None, self.dbos._serializer
            )
            await asyncio.to_thread(
                self.dbos._sys_db.record_get_result,
                self.workflow_id,
                None,
                serialized_e,
                serialization,
            )
            raise
        serialized_r, serialization = serialize_value(r, None, self.dbos._serializer)
        await asyncio.to_thread(
            self.dbos._sys_db.record_get_result,
            self.workflow_id,
            serialized_r,
            None,
            serialization,
        )
        return r

    async def get_status(self) -> WorkflowStatus:
        stat = await asyncio.to_thread(self.dbos.get_workflow_status, self.workflow_id)
        if stat is None:
            raise DBOSNonExistentWorkflowError("target", self.workflow_id)
        return stat


class WorkflowHandleAsyncPolling(Generic[R]):

    def __init__(self, workflow_id: str, dbos: "DBOS"):
        self.workflow_id = workflow_id
        self.dbos = dbos

    def get_workflow_id(self) -> str:
        return self.workflow_id

    async def get_result(
        self, *, polling_interval_sec: float = DEFAULT_POLLING_INTERVAL
    ) -> R:
        try:
            r: R = await self.dbos._sys_db.await_workflow_result_async(
                self.workflow_id,
                polling_interval_sec,
            )
        except Exception as e:
            serialized_e, serialization = serialize_exception(
                e, None, self.dbos._serializer
            )
            await asyncio.to_thread(
                self.dbos._sys_db.record_get_result,
                self.workflow_id,
                None,
                serialized_e,
                serialization,
            )
            raise
        serialized_r, serialization = serialize_value(r, None, self.dbos._serializer)
        await asyncio.to_thread(
            self.dbos._sys_db.record_get_result,
            self.workflow_id,
            serialized_r,
            None,
            serialization,
        )
        return r

    async def get_status(self) -> WorkflowStatus:
        stat = await asyncio.to_thread(self.dbos.get_workflow_status, self.workflow_id)
        if stat is None:
            raise DBOSNonExistentWorkflowError("target", self.workflow_id)
        return stat


class StepOptions(TypedDict, total=False):
    """
    Configuration options for steps.

    Attributes:
        name:
            Optional name for the step.
            If not provided, the function's name will be used.

        retries_allowed:
            Whether the step should be retried on failure.

        interval_seconds:
            Initial delay (in seconds) between retry attempts.

        max_attempts:
            Maximum number of attempts before the step is
            considered failed.

        backoff_rate:
            Multiplier applied to `interval_seconds` after
            each failed attempt (e.g. 2.0 = exponential backoff).
    """

    name: Optional[str]
    retries_allowed: bool
    interval_seconds: float
    max_attempts: int
    backoff_rate: float


DEFAULT_STEP_OPTIONS: StepOptions = {
    "name": None,
    "retries_allowed": False,
    "interval_seconds": 1.0,
    "max_attempts": 3,
    "backoff_rate": 2.0,
}


def normalize_step_options(opts: Optional[StepOptions]) -> StepOptions:
    return {**DEFAULT_STEP_OPTIONS, **(opts or {})}


def _init_workflow(
    dbos: "DBOS",
    ctx: DBOSContext,
    *,
    inputs: WorkflowInputs,
    wf_name: str,
    class_name: Optional[str],
    config_name: Optional[str],
    queue: Optional[str],
    workflow_timeout_ms: Optional[int],
    workflow_deadline_epoch_ms: Optional[int],
    max_recovery_attempts: Optional[int],
    enqueue_options: Optional[EnqueueOptionsInternal],
    is_recovery_request: Optional[bool],
    is_dequeued_request: Optional[bool],
    serialization_type: Optional[WorkflowSerializationFormat],
) -> tuple[WorkflowStatusInternal, bool]:
    wfid = (
        ctx.workflow_id
        if len(ctx.workflow_id) > 0
        else ctx.id_assigned_for_next_workflow
    )

    # If we have a class name, the first arg is the instance and do not serialize
    if class_name is not None and class_name != "":
        inputs = {"args": inputs["args"][1:], "kwargs": inputs["kwargs"]}

    sertype: WorkflowSerializationFormat | None = serialization_type
    if sertype is None or sertype == WorkflowSerializationFormat.DEFAULT:
        sertype = ctx.serialization_type
    serargs, serialization = serialize_args(
        inputs["args"], inputs["kwargs"], sertype, dbos._serializer
    )

    # Initialize a workflow status object from the context
    status: WorkflowStatusInternal = {
        "workflow_uuid": wfid,
        "status": (
            WorkflowStatusString.PENDING.value
            if queue is None
            else WorkflowStatusString.ENQUEUED.value
        ),
        "name": wf_name,
        "class_name": class_name,
        "config_name": config_name,
        "output": None,
        "error": None,
        "app_id": ctx.app_id,
        "app_version": (
            enqueue_options["app_version"]
            if enqueue_options is not None
            and enqueue_options["app_version"] is not None
            else GlobalParams.app_version
        ),
        "executor_id": ctx.executor_id,
        "recovery_attempts": None,
        "authenticated_user": ctx.authenticated_user,
        "authenticated_roles": (
            json.dumps(ctx.authenticated_roles) if ctx.authenticated_roles else None
        ),
        "assumed_role": ctx.assumed_role,
        "queue_name": queue,
        "created_at": None,
        "updated_at": None,
        "workflow_timeout_ms": workflow_timeout_ms,
        "workflow_deadline_epoch_ms": workflow_deadline_epoch_ms,
        "deduplication_id": (
            enqueue_options["deduplication_id"] if enqueue_options is not None else None
        ),
        "priority": (
            (
                enqueue_options["priority"]
                if enqueue_options["priority"] is not None
                else 0
            )
            if enqueue_options is not None
            else 0
        ),
        "inputs": serargs,
        "queue_partition_key": (
            enqueue_options["queue_partition_key"]
            if enqueue_options is not None
            else None
        ),
        "forked_from": None,
        "parent_workflow_id": (
            ctx.parent_workflow_id if len(ctx.parent_workflow_id) > 0 else None
        ),
        "started_at_epoch_ms": None,
        "owner_xid": None,
        "serialization": serialization,
    }

    # Synchronously record the status and inputs for workflows
    try:
        wf_status, workflow_deadline_epoch_ms, should_execute = (
            dbos._sys_db.init_workflow(
                status,
                max_recovery_attempts=max_recovery_attempts,
                is_dequeued_request=is_dequeued_request,
                is_recovery_request=is_recovery_request,
                owner_xid=str(uuid.uuid4()),
            )
        )
    except DBOSQueueDeduplicatedError as e:
        sererr, serialization = serialize_exception(
            e,
            status["serialization"],
            dbos._serializer,
        )
        if ctx.has_parent():
            result: OperationResultInternal = {
                "workflow_uuid": ctx.parent_workflow_id,
                "function_id": ctx.parent_workflow_fid,
                "function_name": wf_name,
                "output": None,
                "error": sererr,
                "serialization": serialization,
                "started_at_epoch_ms": int(time.time() * 1000),
            }
            dbos._sys_db.record_operation_result(result)
        raise

    if should_execute and workflow_deadline_epoch_ms is not None:
        evt = threading.Event()
        dbos.background_thread_stop_events.append(evt)

        def timeout_func() -> None:
            try:
                assert workflow_deadline_epoch_ms is not None
                time_to_wait_sec = (
                    workflow_deadline_epoch_ms - (time.time() * 1000)
                ) / 1000
                if time_to_wait_sec > 0:
                    was_stopped = evt.wait(time_to_wait_sec)
                    if was_stopped:
                        return
                dbos._sys_db.cancel_workflow(wfid)
            except Exception as e:
                dbos.logger.warning(
                    f"Exception in timeout thread for workflow {wfid}: {e}"
                )

        timeout_thread = threading.Thread(target=timeout_func, daemon=True)
        timeout_thread.start()
        dbos._background_threads.append(timeout_thread)

    ctx.workflow_deadline_epoch_ms = workflow_deadline_epoch_ms
    status["workflow_deadline_epoch_ms"] = workflow_deadline_epoch_ms
    status["status"] = wf_status
    return status, should_execute


def _get_wf_invoke_func(
    dbos: "DBOS",
    status: WorkflowStatusInternal,
) -> Callable[[Callable[[], R]], R]:
    def persist(func: Callable[[], R]) -> R:
        if (
            status["status"] == WorkflowStatusString.ERROR.value
            or status["status"] == WorkflowStatusString.SUCCESS.value
        ):
            dbos.logger.debug(
                f"Workflow {status['workflow_uuid']} is already completed with status {status['status']}"
            )
            # Directly return the result if the workflow is already completed
            recorded_result: R = dbos._sys_db.await_workflow_result(
                status["workflow_uuid"], polling_interval=DEFAULT_POLLING_INTERVAL
            )
            return recorded_result
        try:
            if inspect.iscoroutinefunction(func):
                output = dbos._background_event_loop.submit_coroutine(
                    cast(Coroutine[Any, Any, R], func())
                )
            else:
                output = func()

            serval, _serialization = serialize_value_as(
                output, status["serialization"], dbos._serializer
            )
            dbos._sys_db.update_workflow_outcome(
                status["workflow_uuid"],
                WorkflowStatusString.SUCCESS.value,
                output=serval,
            )
            return output
        except DBOSWorkflowConflictIDError:
            # Await the workflow result
            r: R = dbos._sys_db.await_workflow_result(
                status["workflow_uuid"], polling_interval=DEFAULT_POLLING_INTERVAL
            )
            return r
        except DBOSWorkflowCancelledError as error:
            raise DBOSAwaitedWorkflowCancelledError(status["workflow_uuid"])
        except Exception as error:
            dbos._sys_db.update_workflow_outcome(
                status["workflow_uuid"],
                WorkflowStatusString.ERROR.value,
                error=serialize_exception(
                    error,
                    status["serialization"],
                    dbos._serializer,
                )[0],
            )
            raise

    return persist


class ActiveWorkflowById:
    def __init__(self) -> None:
        self._lock = threading.Lock()
        self._m: dict[str, bool] = {}

    def acquire(self, key: str) -> bool:
        """
        Returns is_owner
        """
        with self._lock:
            entry = self._m.get(key)
            if entry is None:
                self._m[key] = True
                return True
            return False

    def release(
        self,
        key: str,
    ) -> None:
        """
        Removes the key when work done
        """
        with self._lock:
            del self._m[key]

    def activeList(self) -> List[str]:
        return list(self._m.keys())


def _execute_workflow_wthread(
    dbos: "DBOS",
    status: WorkflowStatusInternal,
    func: "Callable[P, R]",
    ctx: DBOSContext,
    args: tuple[Any],
    kwargs: dict[str, Any],
) -> R:
    attributes: TracedAttributes = {
        "name": get_dbos_func_name(func),
        "operationType": OperationType.WORKFLOW.value,
    }
    fi = get_func_info(func)
    with EnterDBOSWorkflow(attributes, ctx):
        rr: Optional[str] = check_required_roles(func, fi)
        with DBOSAssumeRole(rr):
            owned = dbos._active_workflows_set.acquire(status["workflow_uuid"])
            try:
                if owned:
                    return _get_wf_invoke_func(dbos, status)(
                        functools.partial(func, *args, **kwargs)
                    )
                else:
                    output: R = dbos._sys_db.await_workflow_result(
                        status["workflow_uuid"],
                        polling_interval=DEFAULT_POLLING_INTERVAL,
                    )
                    return output
            except Exception as e:
                dbos.logger.error(
                    f"Exception encountered in asynchronous workflow:", exc_info=e
                )
                raise
            finally:
                if owned:
                    dbos._active_workflows_set.release(status["workflow_uuid"])


async def _execute_workflow_async(
    dbos: "DBOS",
    status: WorkflowStatusInternal,
    func: "Callable[P, Coroutine[Any, Any, R]]",
    ctx: DBOSContext,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> R:
    attributes: TracedAttributes = {
        "name": get_dbos_func_name(func),
        "operationType": OperationType.WORKFLOW.value,
    }
    fi = get_func_info(func)
    with EnterDBOSWorkflow(attributes, ctx):
        rr: Optional[str] = check_required_roles(func, fi)
        with DBOSAssumeRole(rr):
            owned = dbos._active_workflows_set.acquire(status["workflow_uuid"])
            try:
                if owned:
                    result = Pending[R](functools.partial(func, *args, **kwargs)).then(
                        _get_wf_invoke_func(dbos, status)
                    )
                    return await result()
                else:

                    def fn() -> Any:
                        return dbos._sys_db.await_workflow_result(
                            status["workflow_uuid"],
                            polling_interval=DEFAULT_POLLING_INTERVAL,
                        )

                    return await asyncio.to_thread(fn)
            except Exception as e:
                dbos.logger.error(
                    f"Exception encountered in asynchronous workflow:", exc_info=e
                )
                raise
            finally:
                if owned:
                    dbos._active_workflows_set.release(status["workflow_uuid"])


def execute_workflow_by_id(
    dbos: "DBOS", workflow_id: str, is_recovery: bool, is_dequeue: bool
) -> "WorkflowHandle[Any]":
    status = dbos._sys_db.get_workflow_status(workflow_id)
    if not status:
        raise DBOSRecoveryError(workflow_id, "Workflow status not found")
    try:
        inputs: WorkflowInputs = deserialize_args(
            status["inputs"], status["serialization"], dbos._serializer
        )
    except Exception as deser_error:
        # Mark workflow as ERROR immediately instead of leaving it PENDING for infinite retry
        try:
            error_str = serialize_exception(
                deser_error, status["serialization"], dbos._serializer
            )[0]
        except Exception:
            # Fallback: create a simple error we know can be serialized
            fallback = Exception(f"{type(deser_error).__name__}: {deser_error}")
            error_str = serialize_exception(
                fallback, status["serialization"], dbos._serializer
            )[0]
        dbos._sys_db.update_workflow_outcome(
            workflow_id,
            WorkflowStatusString.ERROR.value,
            error=error_str,
        )
        raise
    wf_func = dbos._registry.workflow_info_map.get(status["name"], None)
    if not wf_func:
        raise DBOSWorkflowFunctionNotFoundError(
            workflow_id,
            f"{status['name']} is not a registered workflow function",
        )
    fi = get_func_info(wf_func)
    if fi is None:
        raise DBOSWorkflowFunctionNotFoundError(
            "<NONE>",
            f"{wf_func.__name__} is not a registered workflow function",
        )
    # Run argument validation if configured on the workflow
    if fi.validate_args is not None and inputs is not None:
        try:
            validated_args, validated_kwargs = fi.validate_args(
                inputs["args"], inputs["kwargs"]
            )
            inputs = {"args": validated_args, "kwargs": validated_kwargs}
        except Exception as val_error:
            try:
                error_str = serialize_exception(
                    val_error, status["serialization"], dbos._serializer
                )[0]
            except Exception:
                # Fallback: create a simple error we know can be serialized
                fallback = Exception(f"{type(val_error).__name__}: {val_error}")
                error_str = serialize_exception(
                    fallback, status["serialization"], dbos._serializer
                )[0]
            dbos._sys_db.update_workflow_outcome(
                workflow_id,
                WorkflowStatusString.ERROR.value,
                error=error_str,
            )
            raise
    # Restore authentication context from the saved workflow status
    recovered_user = status.get("authenticated_user")
    recovered_roles_str = status.get("authenticated_roles")
    recovered_roles = json.loads(recovered_roles_str) if recovered_roles_str else None

    with DBOSContextSetAuth(recovered_user, recovered_roles):
        # If this function belongs to a configured class, add that class instance as its first argument
        if status["config_name"] is not None:
            config_name = status["config_name"]
            class_name = status["class_name"]
            iname = f"{class_name}/{config_name}"
            if iname not in dbos._registry.instance_info_map:
                raise DBOSWorkflowFunctionNotFoundError(
                    workflow_id,
                    f"configured class instance '{iname}' is not registered",
                )
            class_instance = dbos._registry.instance_info_map[iname]
            inputs["args"] = (class_instance,) + inputs["args"]
        # If this function is a class method, add that class object as its first argument
        elif status["class_name"] is not None:
            class_name = status["class_name"]
            if class_name not in dbos._registry.class_info_map:
                raise DBOSWorkflowFunctionNotFoundError(
                    workflow_id,
                    f"class '{class_name}' is not registered",
                )
            class_object = dbos._registry.class_info_map[class_name]
            if fi.func_type != DBOSFuncType.Static:
                inputs["args"] = (class_object,) + inputs["args"]

        with SetWorkflowID(workflow_id):
            if inspect.iscoroutinefunction(wf_func):
                ctx = get_local_dbos_context()
                parent_ctx_copy = copy.copy(ctx)
                child_ctx = DBOSContext.create_start_workflow_child(ctx)
                async_handle = dbos._background_event_loop.submit_coroutine(
                    start_workflow_async(
                        dbos,
                        parent_ctx_copy,
                        child_ctx,
                        wf_func,
                        inputs["args"],
                        inputs["kwargs"],
                        queue_name=status["queue_name"],
                        execute_workflow=True,
                        is_recovery_request=is_recovery,
                        is_dequeued_request=is_dequeue,
                    )
                )
                return WorkflowHandlePolling(async_handle.get_workflow_id(), dbos)
            return start_workflow(
                dbos,
                wf_func,
                inputs["args"],
                inputs["kwargs"],
                queue_name=status["queue_name"],
                execute_workflow=True,
                is_recovery=is_recovery,
                is_dequeued=is_dequeue,
            )


def start_workflow(
    dbos: "DBOS",
    func: "Callable[P, Union[R, Coroutine[Any, Any, R]]]",
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    queue_name: Optional[str] = None,
    execute_workflow: bool = True,
    is_recovery: bool = False,
    is_dequeued: bool = False,
) -> "WorkflowHandle[R]":

    # If the function has a class, add the class object as its first argument
    fself: Optional[object] = None
    if hasattr(func, "__self__"):
        fself = func.__self__
    if fself is not None:
        args = (fself,) + args

    fi = get_func_info(func)
    if fi is None:
        raise DBOSWorkflowFunctionNotFoundError(
            "<NONE>",
            f"{func.__name__} is not a registered workflow function",
        )
    serialization_type = fi.serialization_type
    if serialization_type is None:
        serialization_type = WorkflowSerializationFormat.DEFAULT

    func = cast("Workflow[P, R]", func.__orig_func)  # type: ignore

    inputs: WorkflowInputs = {
        "args": args,
        "kwargs": kwargs,
    }

    local_ctx = get_local_dbos_context()
    workflow_timeout_ms, workflow_deadline_epoch_ms = _get_timeout_deadline(
        local_ctx, queue_name
    )
    workflow_timeout_ms = (
        local_ctx.workflow_timeout_ms if local_ctx is not None else None
    )
    enqueue_options = EnqueueOptionsInternal(
        deduplication_id=local_ctx.deduplication_id if local_ctx is not None else None,
        priority=local_ctx.priority if local_ctx is not None else None,
        app_version=local_ctx.app_version if local_ctx is not None else None,
        queue_partition_key=(
            local_ctx.queue_partition_key if local_ctx is not None else None
        ),
    )
    new_wf_ctx = DBOSContext.create_start_workflow_child(local_ctx)
    new_child_workflow_id = new_wf_ctx.id_assigned_for_next_workflow

    if new_wf_ctx.has_parent():
        recorded_result = dbos._sys_db.check_operation_execution(
            new_wf_ctx.parent_workflow_id,
            new_wf_ctx.parent_workflow_fid,
            get_dbos_func_name(func),
        )
        if recorded_result and recorded_result["error"]:
            e: Exception = deserialize_exception(
                recorded_result["error"],
                recorded_result["serialization"],
                dbos._sys_db.serializer,
            )
            raise e
        elif recorded_result and recorded_result["child_workflow_id"]:
            return WorkflowHandlePolling(recorded_result["child_workflow_id"], dbos)

    status, should_execute = _init_workflow(
        dbos,
        new_wf_ctx,
        inputs=inputs,
        wf_name=get_dbos_func_name(func),
        class_name=get_dbos_class_name(fi, func, args),
        config_name=get_config_name(fi, func, args),
        queue=queue_name,
        workflow_timeout_ms=workflow_timeout_ms,
        workflow_deadline_epoch_ms=workflow_deadline_epoch_ms,
        max_recovery_attempts=fi.max_recovery_attempts,
        enqueue_options=enqueue_options,
        is_recovery_request=is_recovery,
        is_dequeued_request=is_dequeued,
        serialization_type=serialization_type,
    )

    if status["serialization"] == DBOSPortableJSON.name():
        serialization_type = WorkflowSerializationFormat.PORTABLE
    new_wf_ctx.serialization_type = serialization_type

    wf_status = status["status"]
    if new_wf_ctx.has_parent():
        dbos._sys_db.record_child_workflow(
            new_wf_ctx.parent_workflow_id,
            new_child_workflow_id,
            new_wf_ctx.parent_workflow_fid,
            get_dbos_func_name(func),
        )

    if (
        not execute_workflow
        or not should_execute
        or wf_status == WorkflowStatusString.ERROR.value
        or wf_status == WorkflowStatusString.SUCCESS.value
    ):
        return WorkflowHandlePolling(new_child_workflow_id, dbos)

    future = dbos._executor.submit(
        cast(Callable[..., R], _execute_workflow_wthread),
        dbos,
        status,
        func,
        new_wf_ctx,
        args,
        kwargs,
    )
    return WorkflowHandleFuture(new_child_workflow_id, future, dbos)


async def start_workflow_async(
    dbos: "DBOS",
    local_ctx: Optional[DBOSContext],
    new_wf_ctx: DBOSContext,
    func: "Callable[P, Coroutine[Any, Any, R]]",
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    queue_name: Optional[str] = None,
    execute_workflow: bool = True,
    is_recovery_request: bool = False,
    is_dequeued_request: bool = False,
) -> "WorkflowHandleAsync[R]":
    # If the function has a class, add the class object as its first argument
    fself: Optional[object] = None
    if hasattr(func, "__self__"):
        fself = func.__self__
    if fself is not None:
        args = (fself,) + args

    fi = get_func_info(func)
    if fi is None:
        raise DBOSWorkflowFunctionNotFoundError(
            "<NONE>",
            f"{func.__name__} is not a registered workflow function",
        )
    serialization_type = fi.serialization_type
    if serialization_type is None:
        serialization_type = WorkflowSerializationFormat.DEFAULT

    func = cast("Workflow[P, R]", func.__orig_func)  # type: ignore

    inputs: WorkflowInputs = {
        "args": args,
        "kwargs": kwargs,
    }

    workflow_timeout_ms, workflow_deadline_epoch_ms = _get_timeout_deadline(
        local_ctx, queue_name
    )
    enqueue_options = EnqueueOptionsInternal(
        deduplication_id=local_ctx.deduplication_id if local_ctx is not None else None,
        priority=local_ctx.priority if local_ctx is not None else None,
        app_version=local_ctx.app_version if local_ctx is not None else None,
        queue_partition_key=(
            local_ctx.queue_partition_key if local_ctx is not None else None
        ),
    )
    new_child_workflow_id = new_wf_ctx.id_assigned_for_next_workflow

    if new_wf_ctx.has_parent():
        recorded_result = await asyncio.to_thread(
            dbos._sys_db.check_operation_execution,
            new_wf_ctx.parent_workflow_id,
            new_wf_ctx.parent_workflow_fid,
            get_dbos_func_name(func),
        )
        if recorded_result and recorded_result["error"]:
            e: Exception = deserialize_exception(
                recorded_result["error"],
                recorded_result["serialization"],
                dbos._sys_db.serializer,
            )
            raise e
        elif recorded_result and recorded_result["child_workflow_id"]:
            return WorkflowHandleAsyncPolling(
                recorded_result["child_workflow_id"], dbos
            )

    status, should_execute = await asyncio.to_thread(
        _init_workflow,
        dbos,
        new_wf_ctx,
        inputs=inputs,
        wf_name=get_dbos_func_name(func),
        class_name=get_dbos_class_name(fi, func, args),
        config_name=get_config_name(fi, func, args),
        queue=queue_name,
        workflow_timeout_ms=workflow_timeout_ms,
        workflow_deadline_epoch_ms=workflow_deadline_epoch_ms,
        max_recovery_attempts=fi.max_recovery_attempts,
        enqueue_options=enqueue_options,
        is_recovery_request=is_recovery_request,
        is_dequeued_request=is_dequeued_request,
        serialization_type=serialization_type,
    )

    if status["serialization"] == DBOSPortableJSON.name():
        serialization_type = WorkflowSerializationFormat.PORTABLE
    new_wf_ctx.serialization_type = serialization_type

    if new_wf_ctx.has_parent():
        await asyncio.to_thread(
            dbos._sys_db.record_child_workflow,
            new_wf_ctx.parent_workflow_id,
            new_child_workflow_id,
            new_wf_ctx.parent_workflow_fid,
            get_dbos_func_name(func),
        )

    wf_status = status["status"]

    if (
        not execute_workflow
        or not should_execute
        or wf_status == WorkflowStatusString.ERROR.value
        or wf_status == WorkflowStatusString.SUCCESS.value
    ):
        return WorkflowHandleAsyncPolling(new_child_workflow_id, dbos)

    coro = _execute_workflow_async(dbos, status, func, new_wf_ctx, args, kwargs)
    # Shield the workflow task from cancellation
    task = asyncio.shield(asyncio.create_task(coro))
    return WorkflowHandleAsyncTask(new_child_workflow_id, task, dbos)


if sys.version_info < (3, 12):

    def _mark_coroutine(func: Callable[P, R]) -> Callable[P, R]:
        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> R:
            return await func(*args, **kwargs)  # type: ignore

        return async_wrapper  # type: ignore

else:

    def _mark_coroutine(func: Callable[P, R]) -> Callable[P, R]:
        inspect.markcoroutinefunction(func)

        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> R:
            return await func(*args, **kwargs)  # type: ignore

        return async_wrapper  # type: ignore


def workflow_wrapper(
    dbosreg: "DBOSRegistry",
    func: Callable[P, R],
    max_recovery_attempts: Optional[int] = DEFAULT_MAX_RECOVERY_ATTEMPTS,
    *,
    serialization_type: WorkflowSerializationFormat = WorkflowSerializationFormat.DEFAULT,
    validate_args: Optional["ValidateArgsCallable"] = None,
) -> Callable[P, R]:
    func.__orig_func = func  # type: ignore

    fi = get_or_create_func_info(func)
    fi.max_recovery_attempts = max_recovery_attempts
    fi.serialization_type = serialization_type
    fi.validate_args = validate_args

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> R:
        fi = get_func_info(func)
        assert fi is not None
        if dbosreg.dbos is None:
            raise DBOSException(
                f"Function {get_dbos_func_name(func)} invoked before DBOS initialized"
            )
        dbos = dbosreg.dbos

        rr: Optional[str] = check_required_roles(func, fi)
        attributes: TracedAttributes = {
            "name": get_dbos_func_name(func),
            "operationType": OperationType.WORKFLOW.value,
        }
        inputs: WorkflowInputs = {
            "args": args,
            "kwargs": kwargs,
        }
        cctx = get_local_dbos_context()
        newwfctx = DBOSContext.create_start_workflow_child(cctx)
        resctx: Optional[DBOSContext] = None
        if cctx is not None and cctx.is_workflow():
            resctx = cctx.snapshot_step_ctx(reserve_sleep_id=False)
        workflow_timeout_ms, workflow_deadline_epoch_ms = _get_timeout_deadline(
            cctx, queue=None
        )

        wfOutcome = Outcome[R].make(functools.partial(func, *args, **kwargs))

        workflow_id = None

        def init_wf() -> Callable[[Callable[[], R]], R]:

            def recorded_result(
                c_wfid: str, dbos: "DBOS"
            ) -> Callable[[Callable[[], R]], R]:
                def recorded_result_inner(func: Callable[[], R]) -> R:
                    r: R = dbos._sys_db.await_workflow_result(
                        c_wfid, polling_interval=DEFAULT_POLLING_INTERVAL
                    )
                    return r

                return recorded_result_inner

            nonlocal workflow_id
            workflow_id = newwfctx.workflow_id

            if newwfctx.has_parent():
                r = dbos._sys_db.check_operation_execution(
                    newwfctx.parent_workflow_id,
                    newwfctx.parent_workflow_fid,
                    get_dbos_func_name(func),
                )
                if r and r["error"]:
                    e: Exception = deserialize_exception(
                        r["error"], r["serialization"], dbos._sys_db.serializer
                    )
                    raise e
                elif r and r["child_workflow_id"]:
                    return recorded_result(r["child_workflow_id"], dbos)

            status, should_execute = _init_workflow(
                dbos,
                newwfctx,
                inputs=inputs,
                wf_name=get_dbos_func_name(func),
                class_name=get_dbos_class_name(fi, func, args),
                config_name=get_config_name(fi, func, args),
                queue=None,
                workflow_timeout_ms=workflow_timeout_ms,
                workflow_deadline_epoch_ms=workflow_deadline_epoch_ms,
                max_recovery_attempts=max_recovery_attempts,
                enqueue_options=None,
                is_recovery_request=False,
                is_dequeued_request=False,
                serialization_type=fi.serialization_type,
            )

            def get_recorded_result(_func: Callable[[], R]) -> R:
                return cast(
                    R,
                    dbos._sys_db.await_workflow_result(
                        status["workflow_uuid"],
                        polling_interval=DEFAULT_POLLING_INTERVAL,
                    ),
                )

            # TODO: maybe modify the parameters if they've been changed by `_init_workflow`
            dbos.logger.debug(
                f"Running workflow, id: {newwfctx.workflow_id}, name: {get_dbos_func_name(func)}"
            )

            if newwfctx.has_parent():
                dbos._sys_db.record_child_workflow(
                    newwfctx.parent_workflow_id,
                    newwfctx.workflow_id,
                    newwfctx.parent_workflow_fid,
                    get_dbos_func_name(func),
                )

            if should_execute:
                return _get_wf_invoke_func(dbos, status)
            else:
                dbos.logger.debug(
                    f"Workflow {status['workflow_uuid']} already run with status {status['status']}"
                )
                return get_recorded_result

        def record_get_result(func: Callable[[], R]) -> R:
            """
            If a child workflow is invoked synchronously, this records the implicit "getResult" where the
            parent retrieves the child's output. It executes in the CALLER'S context, not the workflow's.
            """
            try:
                r = func()
            except Exception as e:
                serialized_e, serialization = serialize_exception(
                    e, None, dbos._serializer
                )
                assert workflow_id is not None
                dbos._sys_db.record_get_result(
                    workflow_id, None, serialized_e, serialization, resctx
                )
                raise
            serialized_r, serialization = serialize_value(r, None, dbos._serializer)
            assert workflow_id is not None
            dbos._sys_db.record_get_result(
                workflow_id, serialized_r, None, serialization, resctx
            )
            return r

        outcome = (
            wfOutcome.wrap(init_wf, dbos=dbos)
            .also(DBOSAssumeRole(rr))
            .also(EnterDBOSWorkflow(attributes, newwfctx))
            .then(record_get_result, dbos=dbos)
        )
        return outcome()  # type: ignore

    return _mark_coroutine(wrapper) if inspect.iscoroutinefunction(func) else wrapper


def decorate_workflow(
    reg: "DBOSRegistry",
    name: Optional[str],
    max_recovery_attempts: Optional[int],
    *,
    serialization_type: Optional[WorkflowSerializationFormat] = None,
    validate_args: Optional["ValidateArgsCallable"] = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    if serialization_type is None:
        serialization_type = WorkflowSerializationFormat.DEFAULT

    def _workflow_decorator(func: Callable[P, R]) -> Callable[P, R]:
        resolved_validate_args = validate_args
        # If pydantic_args_validator sentinel is passed, build a real validator
        # from the function's type hints at decoration time
        if resolved_validate_args is not None:
            from ._validation import pydantic_args_validator

            if resolved_validate_args is pydantic_args_validator:
                from ._validation import make_pydantic_args_validator

                resolved_validate_args = make_pydantic_args_validator(func)
        wrapped_func = workflow_wrapper(
            reg,
            func,
            max_recovery_attempts,
            serialization_type=serialization_type,
            validate_args=resolved_validate_args,
        )
        func_name = name if name is not None else func.__qualname__
        set_dbos_func_name(func, func_name)
        set_dbos_func_name(wrapped_func, func_name)
        reg.register_wf_function(func_name, wrapped_func, "workflow")
        return wrapped_func

    return _workflow_decorator


def decorate_transaction(
    dbosreg: "DBOSRegistry", name: Optional[str], isolation_level: "IsolationLevel"
) -> Callable[[F], F]:
    def decorator(func: F) -> F:

        transaction_name = name if name is not None else func.__qualname__

        def invoke_tx(*args: Any, **kwargs: Any) -> Any:
            if dbosreg.dbos is None:
                raise DBOSException(
                    f"Function {transaction_name} invoked before DBOS initialized"
                )
            dbos = dbosreg.dbos

            try:
                asyncio.get_running_loop()
            except RuntimeError:
                pass
            else:
                dbosreg.dbos.logger.warning(
                    f"Transaction {transaction_name} was called while an event loop is running. Invoke transactions from an async context using asyncio.to_thread to avoid blocking the event loop."
                )

            assert (
                dbos._app_db
            ), "Transactions can only be used if DBOS is configured with an application_database_url"
            with dbos._app_db.sessionmaker() as session:
                attributes: TracedAttributes = {
                    "name": transaction_name,
                    "operationType": OperationType.TRANSACTION.value,
                }
                with EnterDBOSTransaction(session, attributes=attributes):
                    ctx = assert_current_dbos_context()
                    # Check if the step record for this transaction exists
                    recorded_step_output = dbos._sys_db.check_operation_execution(
                        ctx.workflow_id, ctx.function_id, transaction_name
                    )
                    if recorded_step_output:
                        dbos.logger.debug(
                            f"Replaying transaction, id: {ctx.function_id}, name: {attributes['name']}"
                        )
                        if recorded_step_output["error"]:
                            step_error: Exception = deserialize_exception(
                                recorded_step_output["error"],
                                recorded_step_output["serialization"],
                                dbos._serializer,
                            )
                            raise step_error
                        elif recorded_step_output["output"]:
                            return deserialize_value(
                                recorded_step_output["output"],
                                recorded_step_output["child_workflow_id"],
                                dbos._serializer,
                            )
                        else:
                            raise Exception("Output and error are both None")

                    txn_output: TransactionResultInternal = {
                        "workflow_uuid": ctx.workflow_id,
                        "function_id": ctx.function_id,
                        "output": None,
                        "error": None,
                        "serialization": None,
                        "txn_snapshot": "",  # TODO: add actual snapshot
                        "executor_id": None,
                        "txn_id": None,
                        "function_name": transaction_name,
                    }
                    step_output: OperationResultInternal = {
                        "workflow_uuid": ctx.workflow_id,
                        "function_id": ctx.function_id,
                        "function_name": transaction_name,
                        "output": None,
                        "error": None,
                        "serialization": None,
                        "started_at_epoch_ms": int(time.time() * 1000),
                    }
                    retry_wait_seconds = 0.001
                    backoff_factor = 1.5
                    max_retry_wait_seconds = 2.0
                    while True:
                        has_recorded_error = False
                        txn_error: Optional[Exception] = None
                        try:
                            with session.begin():
                                # This must be the first statement in the transaction!
                                session.connection(
                                    execution_options={
                                        "isolation_level": isolation_level
                                    }
                                )
                                # Check recorded output for OAOO
                                recorded_output = (
                                    ApplicationDatabase.check_transaction_execution(
                                        session,
                                        ctx.workflow_id,
                                        ctx.function_id,
                                        transaction_name,
                                    )
                                )
                                if recorded_output:
                                    dbos.logger.debug(
                                        f"Replaying transaction, id: {ctx.function_id}, name: {attributes['name']}"
                                    )
                                    if recorded_output["error"]:
                                        deserialized_error: Exception = (
                                            deserialize_exception(
                                                recorded_output["error"],
                                                recorded_output["serialization"],
                                                dbos._serializer,
                                            )
                                        )
                                        has_recorded_error = True
                                        step_output["error"] = recorded_output["error"]
                                        step_output["serialization"] = recorded_output[
                                            "serialization"
                                        ]
                                        dbos._sys_db.record_operation_result(
                                            step_output
                                        )
                                        raise deserialized_error
                                    elif recorded_output["output"]:
                                        step_output["output"] = recorded_output[
                                            "output"
                                        ]
                                        step_output["serialization"] = recorded_output[
                                            "serialization"
                                        ]
                                        dbos._sys_db.record_operation_result(
                                            step_output
                                        )
                                        return deserialize_value(
                                            recorded_output["output"],
                                            recorded_output["serialization"],
                                            dbos._serializer,
                                        )
                                    else:
                                        raise Exception(
                                            "Output and error are both None"
                                        )
                                else:
                                    dbos.logger.debug(
                                        f"Running transaction, id: {ctx.function_id}, name: {attributes['name']}"
                                    )

                                output = func(*args, **kwargs)
                                serialized_r, serialization = serialize_value(
                                    output, None, dbos._serializer
                                )
                                txn_output["output"] = serialized_r
                                txn_output["serialization"] = serialization
                                assert (
                                    ctx.sql_session is not None
                                ), "Cannot find a database connection"
                                dbos._app_db.record_transaction_output(
                                    ctx.sql_session, txn_output
                                )
                                break
                        except DBAPIError as dbapi_error:
                            if retriable_postgres_exception(
                                dbapi_error
                            ) or dbos._app_db._is_serialization_error(dbapi_error):
                                # Retry on serialization failure
                                span = ctx.get_current_dbos_span()
                                if span:
                                    span.add_event(
                                        "Transaction Failure",
                                        {"retry_wait_seconds": retry_wait_seconds},
                                    )
                                time.sleep(retry_wait_seconds)
                                retry_wait_seconds = min(
                                    retry_wait_seconds * backoff_factor,
                                    max_retry_wait_seconds,
                                )
                                continue
                            txn_error = dbapi_error
                            raise
                        except InvalidRequestError as invalid_request_error:
                            dbos.logger.error(
                                f"InvalidRequestError in transaction {transaction_name} \033[1m Hint: Do not call commit() or rollback() within a DBOS transaction.\033[0m"
                            )
                            txn_error = invalid_request_error
                            raise
                        except DBOSUnexpectedStepError:
                            raise
                        except Exception as error:
                            txn_error = error
                            raise
                        finally:
                            # Don't record the error if it was already recorded
                            if txn_error and not has_recorded_error:
                                serialized_e, serialization = serialize_exception(
                                    txn_error, None, dbos._serializer
                                )
                                step_output["error"] = txn_output["error"] = (
                                    serialized_e
                                )
                                step_output["serialization"] = txn_output[
                                    "serialization"
                                ] = serialization
                                dbos._app_db.record_transaction_error(txn_output)
                                dbos._sys_db.record_operation_result(step_output)
            serialized_r, serialization = serialize_value(
                output, None, dbos._serializer
            )
            step_output["output"] = serialized_r
            step_output["serialization"] = serialization
            dbos._sys_db.record_operation_result(step_output)
            return output

        if inspect.iscoroutinefunction(func):
            raise DBOSException(
                f"Function {transaction_name} is a coroutine function, but DBOS.transaction does not support coroutine functions"
            )

        fi = get_or_create_func_info(func)

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            rr: Optional[str] = check_required_roles(func, fi)
            # Entering transaction is allowed:
            #  In a workflow (that is not in a step already)
            #  Not in a workflow (we will start the single op workflow)
            ctx = get_local_dbos_context()
            if ctx and ctx.is_within_workflow():
                assert (
                    ctx.is_workflow()
                ), "Transactions must be called from within workflows"
                with DBOSAssumeRole(rr):
                    return invoke_tx(*args, **kwargs)
            else:
                tempwf = dbosreg.workflow_info_map.get("<temp>." + transaction_name)
                assert tempwf
                return tempwf(*args, **kwargs)

        set_dbos_func_name(func, transaction_name)
        set_dbos_func_name(wrapper, transaction_name)

        def temp_wf(*args: Any, **kwargs: Any) -> Any:
            return wrapper(*args, **kwargs)

        wrapped_wf = workflow_wrapper(dbosreg, temp_wf)
        set_dbos_func_name(temp_wf, "<temp>." + transaction_name)
        set_dbos_func_name(wrapped_wf, "<temp>." + transaction_name)
        set_temp_workflow_type(temp_wf, "transaction")
        dbosreg.register_wf_function(
            get_dbos_func_name(temp_wf), wrapped_wf, "transaction"
        )
        wrapper.__orig_func = temp_wf  # type: ignore
        set_func_info(wrapped_wf, get_or_create_func_info(func))
        set_func_info(temp_wf, get_or_create_func_info(func))

        return cast(F, wrapper)

    return decorator


def invoke_step(
    dbos: "DBOS",
    step_ctx: DBOSContext,
    func: Callable[P, Coroutine[Any, Any, R]] | Callable[P, R],
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    *,
    step_name: str,
    retries_allowed: bool,
    interval_seconds: float,
    max_attempts: int,
    backoff_rate: float,
) -> R | Coroutine[Any, Any, R]:
    attributes: TracedAttributes = {
        "name": step_name,
        "operationType": OperationType.STEP.value,
    }

    step_start_time = int(time.time() * 1000)
    attempts = max_attempts if retries_allowed else 1
    max_retry_interval_seconds: float = 3600  # 1 Hour

    def on_exception(attempt: int, error: BaseException) -> float:
        dbos.logger.warning(
            f"Step being automatically retried (attempt {attempt + 1} of {attempts})",
            exc_info=error,
        )
        ctx = assert_current_dbos_context()
        span = ctx.get_current_dbos_span()
        if span:
            span.add_event(
                f"Step attempt {attempt} failed",
                {
                    "error": str(error),
                    "retryIntervalSeconds": interval_seconds,
                },
            )
        return min(
            interval_seconds * (backoff_rate**attempt),
            max_retry_interval_seconds,
        )

    def record_step_result(func: Callable[[], R]) -> R:
        ctx = assert_current_dbos_context()
        step_output: OperationResultInternal = {
            "workflow_uuid": ctx.workflow_id,
            "function_id": ctx.function_id,
            "function_name": step_name,
            "output": None,
            "error": None,
            "serialization": None,
            "started_at_epoch_ms": step_start_time,
        }

        try:
            output = func()
        except Exception as error:
            serialized_e, serialization = serialize_exception(
                error, None, dbos._serializer
            )
            step_output["error"] = serialized_e
            step_output["serialization"] = serialization
            dbos._sys_db.record_operation_result(step_output)
            raise
        serialized_r, serialization = serialize_value(output, None, dbos._serializer)
        step_output["output"] = serialized_r
        step_output["serialization"] = serialization
        dbos._sys_db.record_operation_result(step_output)
        return output

    def check_existing_result() -> Union[NoResult, R]:
        ctx = assert_current_dbos_context()
        recorded_output = dbos._sys_db.check_operation_execution(
            ctx.workflow_id, ctx.function_id, step_name
        )
        if recorded_output:
            dbos.logger.debug(
                f"Replaying step, id: {ctx.function_id}, name: {attributes['name']}"
            )
            if recorded_output["error"] is not None:
                deserialized_error: Exception = deserialize_exception(
                    recorded_output["error"],
                    recorded_output["serialization"],
                    dbos._serializer,
                )
                raise deserialized_error
            elif recorded_output["output"] is not None:
                return cast(
                    R,
                    deserialize_value(
                        recorded_output["output"],
                        recorded_output["serialization"],
                        dbos._serializer,
                    ),
                )
            else:
                raise Exception("Output and error are both None")
        else:
            dbos.logger.debug(
                f"Running step, id: {ctx.function_id}, name: {attributes['name']}"
            )
            return NoResult()

    stepOutcome = Outcome[R].make(functools.partial(func, *args, **kwargs))
    if retries_allowed:
        stepOutcome = stepOutcome.retry(
            max_attempts,
            on_exception,
            lambda i, e: DBOSMaxStepRetriesExceeded(step_name, i, e),
        )

    outcome = (
        stepOutcome.then(record_step_result)
        .intercept(check_existing_result, dbos=dbos)
        .also(EnterDBOSStepCtx(attributes, step_ctx))
    )
    return outcome()


def run_step(
    dbos: "DBOS",
    func: Callable[P, Coroutine[Any, Any, R]] | Callable[P, R],
    options: StepOptions,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> R:
    options = normalize_step_options(options)
    # If the step is called from a workflow, run it as a step.
    # Otherwise, run it as a normal function.
    ctx = get_local_dbos_context()
    if ctx and ctx.is_workflow():
        outcome = invoke_step(
            dbos,
            ctx.snapshot_step_ctx(),
            func,
            args,
            kwargs,
            step_name=options["name"] if options["name"] else func.__qualname__,
            retries_allowed=options["retries_allowed"],
            interval_seconds=options["interval_seconds"],
            max_attempts=options["max_attempts"],
            backoff_rate=options["backoff_rate"],
        )
        if inspect.iscoroutinefunction(func):
            return dbos._background_event_loop.submit_coroutine(
                cast(Coroutine[Any, Any, R], outcome)
            )
        else:
            return cast(R, outcome)
    else:
        if inspect.iscoroutinefunction(func):

            async def runfunc() -> R:
                return await cast(Callable[P, Coroutine[Any, Any, R]], func)(
                    *args, **kwargs
                )

            return dbos._background_event_loop.submit_coroutine(runfunc())
        else:
            return cast(Callable[P, R], func)(*args, **kwargs)


async def run_step_async(
    dbos: "DBOS",
    step_ctx: Optional[DBOSContext],
    func: Callable[P, Coroutine[Any, Any, R]] | Callable[P, R],
    options: StepOptions,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> R:
    # If the step is called from a workflow, run it as a step.
    # Otherwise, run it as a normal function.
    options = normalize_step_options(options)
    if step_ctx and step_ctx.is_workflow():
        outcome = invoke_step(
            dbos,
            step_ctx,
            func,
            args,
            kwargs,
            step_name=options["name"] if options["name"] else func.__qualname__,
            retries_allowed=options["retries_allowed"],
            interval_seconds=options["interval_seconds"],
            max_attempts=options["max_attempts"],
            backoff_rate=options["backoff_rate"],
        )
        if inspect.iscoroutinefunction(func):
            return await cast(Coroutine[Any, Any, R], outcome)
        else:
            return cast(R, outcome)
    else:
        if inspect.iscoroutinefunction(func):
            return await cast(Callable[P, Coroutine[Any, Any, R]], func)(
                *args, **kwargs
            )
        else:
            return await asyncio.to_thread(
                lambda: cast(Callable[P, R], func)(*args, **kwargs)
            )


def decorate_step(
    dbosreg: "DBOSRegistry",
    *,
    name: Optional[str],
    retries_allowed: bool,
    interval_seconds: float,
    max_attempts: int,
    backoff_rate: float,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    def decorator(func: Callable[P, R]) -> Callable[P, R]:

        step_name = name if name is not None else func.__qualname__

        fi = get_or_create_func_info(func)

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            # If the step is called from a workflow, run it as a step.
            # Otherwise, run it as a normal function.
            ctx = get_local_dbos_context()
            if ctx and ctx.is_workflow():
                if dbosreg.dbos is None:
                    raise DBOSException(
                        f"Function {step_name} invoked before DBOS initialized"
                    )
                rr: Optional[str] = check_required_roles(func, fi)
                with DBOSAssumeRole(rr):
                    return invoke_step(
                        dbosreg.dbos,
                        ctx.snapshot_step_ctx(),
                        func,
                        args,
                        kwargs,
                        step_name=step_name,
                        retries_allowed=retries_allowed,
                        interval_seconds=interval_seconds,
                        max_attempts=max_attempts,
                        backoff_rate=backoff_rate,
                    )
            else:
                return func(*args, **kwargs)

        wrapper = (
            _mark_coroutine(wrapper) if inspect.iscoroutinefunction(func) else wrapper  # type: ignore
        )

        set_dbos_func_name(func, step_name)
        set_dbos_func_name(wrapper, step_name)

        def temp_wf_sync(*args: Any, **kwargs: Any) -> Any:
            return wrapper(*args, **kwargs)

        async def temp_wf_async(*args: Any, **kwargs: Any) -> Any:
            return await wrapper(*args, **kwargs)

        temp_wf = temp_wf_async if inspect.iscoroutinefunction(func) else temp_wf_sync
        wrapped_wf = workflow_wrapper(dbosreg, temp_wf)
        set_dbos_func_name(temp_wf, "<temp>." + step_name)
        set_dbos_func_name(wrapped_wf, "<temp>." + step_name)
        set_temp_workflow_type(temp_wf, "step")
        dbosreg.register_wf_function(get_dbos_func_name(temp_wf), wrapped_wf, "step")
        wrapper.__orig_func = temp_wf  # type: ignore
        set_func_info(wrapped_wf, get_or_create_func_info(func))
        set_func_info(temp_wf, get_or_create_func_info(func))

        return cast(Callable[P, R], wrapper)

    return decorator


def send(
    dbos: "DBOS",
    cur_ctx: Optional["DBOSContext"],
    destination_id: str,
    message: Any,
    topic: Optional[str] = None,
    *,
    serialization_type: Optional[WorkflowSerializationFormat],
    idempotency_key: Optional[str] = None,
) -> None:
    if (
        serialization_type is None
        or serialization_type == WorkflowSerializationFormat.DEFAULT
    ):
        serialization_type = (
            cur_ctx.serialization_type
            if cur_ctx is not None
            else WorkflowSerializationFormat.DEFAULT
        )

    def do_send(destination_id: str, message: Any, topic: Optional[str]) -> None:
        assert cur_ctx is not None
        attributes: TracedAttributes = {
            "name": "send",
        }
        with EnterDBOSStepCtx(attributes, cur_ctx) as ctx:
            dbos._sys_db.send(
                ctx.workflow_id,
                ctx.curr_step_function_id,
                destination_id,
                message,
                topic,
                serialization_type=serialization_type,
                message_uuid=idempotency_key,
            )

    if cur_ctx and cur_ctx.is_workflow():
        return do_send(destination_id, message, topic)
    else:
        dbos._sys_db.send_direct(
            destination_id,
            message,
            topic,
            message_uuid=idempotency_key,
            serialization_type=serialization_type,
        )


def set_event(
    dbos: "DBOS",
    cur_ctx: Optional["DBOSContext"],
    key: str,
    value: Any,
    *,
    serialization_type: WorkflowSerializationFormat,
) -> None:
    if (
        serialization_type is None
        or serialization_type == WorkflowSerializationFormat.DEFAULT
    ):
        serialization_type = (
            cur_ctx.serialization_type
            if cur_ctx is not None
            else WorkflowSerializationFormat.DEFAULT
        )

    if cur_ctx is not None:
        if cur_ctx.is_workflow():
            # If called from a workflow function, run as a step
            attributes: TracedAttributes = {
                "name": "set_event",
            }
            with EnterDBOSStepCtx(attributes, cur_ctx) as ctx:
                dbos._sys_db.set_event_from_workflow(
                    ctx.workflow_id,
                    ctx.curr_step_function_id,
                    key,
                    value,
                    serialization_type=serialization_type,
                )
        elif cur_ctx.is_step():
            dbos._sys_db.set_event_from_step(
                cur_ctx.workflow_id,
                cur_ctx.curr_step_function_id,
                key,
                value,
                serialization_type=serialization_type,
            )
        else:
            raise DBOSException(
                "set_event() must be called from within a workflow or step"
            )
    else:
        raise DBOSException("set_event() must be called from within a workflow or step")


def record_sleep(
    dbos: "DBOS", cur_ctx: Optional["DBOSContext"], seconds: float
) -> float:
    if cur_ctx is not None:
        # Must call it within a workflow
        assert cur_ctx.is_workflow(), "sleep() must be called from within a workflow"
        attributes: TracedAttributes = {
            "name": "sleep",
        }
        with EnterDBOSStepCtx(attributes, cur_ctx) as ctx:
            return dbos._sys_db.record_sleep(
                ctx.workflow_id, ctx.curr_step_function_id, seconds
            )
    else:
        # Cannot call it from outside of a workflow
        raise DBOSException("sleep() must be called from within a workflow")


def write_stream(
    dbos: "DBOS",
    step_ctx: Optional["DBOSContext"],
    key: str,
    value: Any,
    *,
    serialization_type: WorkflowSerializationFormat,
) -> None:
    if (
        serialization_type is None
        or serialization_type == WorkflowSerializationFormat.DEFAULT
    ):
        serialization_type = (
            step_ctx.serialization_type
            if step_ctx is not None
            else WorkflowSerializationFormat.DEFAULT
        )

    if step_ctx is not None:
        # Must call it within a workflow
        if step_ctx.is_workflow():
            attributes: TracedAttributes = {
                "name": "write_stream",
            }
            with EnterDBOSStepCtx(attributes, step_ctx) as ctx:
                dbos._sys_db.write_stream_from_workflow(
                    ctx.workflow_id,
                    ctx.function_id,
                    key,
                    value,
                    serialization_type=serialization_type,
                )
        elif step_ctx.is_step():
            dbos._sys_db.write_stream_from_step(
                step_ctx.workflow_id,
                step_ctx.function_id,
                key,
                value,
                serialization_type=serialization_type,
            )
        else:
            raise DBOSException(
                "write_stream() must be called from within a workflow or step"
            )
    else:
        # Cannot call it from outside of a workflow
        raise DBOSException(
            "write_stream() must be called from within a workflow or step"
        )


def close_stream(dbos: "DBOS", step_ctx: Optional["DBOSContext"], key: str) -> None:
    if step_ctx is not None:
        # Must call it within a workflow
        if step_ctx.is_workflow():
            attributes: TracedAttributes = {
                "name": "close_stream",
            }
            with EnterDBOSStepCtx(attributes, step_ctx) as ctx:
                dbos._sys_db.close_stream(ctx.workflow_id, ctx.function_id, key)
        else:
            raise DBOSException("close_stream() must be called from within a workflow")
    else:
        # Cannot call it from outside of a workflow
        raise DBOSException("close_stream() must be called from within a workflow")


def _get_timeout_deadline(
    ctx: Optional[DBOSContext], queue: Optional[str]
) -> tuple[Optional[int], Optional[int]]:
    if ctx is None:
        return None, None
    # If a timeout is explicitly specified, use it over any propagated deadline
    if ctx.workflow_timeout_ms:
        if queue:
            # Queued workflows are assigned a deadline on dequeue
            return ctx.workflow_timeout_ms, None
        else:
            # Otherwise, compute the deadline immediately
            return (
                ctx.workflow_timeout_ms,
                int(time.time() * 1000) + ctx.workflow_timeout_ms,
            )
    # Otherwise, return the propagated deadline, if any
    else:
        return None, ctx.workflow_deadline_epoch_ms



import copy
import random
import threading
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Optional, TypedDict

from psycopg import errors
from sqlalchemy.exc import OperationalError

from dbos._context import DBOSContext, get_local_dbos_context
from dbos._logger import dbos_logger
from dbos._utils import INTERNAL_QUEUE_NAME, GlobalParams

from ._core import P, R, execute_workflow_by_id, start_workflow, start_workflow_async

if TYPE_CHECKING:
    from ._dbos import DBOS, WorkflowHandle, WorkflowHandleAsync


class QueueRateLimit(TypedDict):
    """
    Limit the maximum number of workflows from this queue that can be started in a given period.

    If the limit is 5 and the period is 10, no more than 5 functions can be
    started per 10 seconds.
    """

    limit: int
    period: float


class Queue:
    """
    Workflow queue.

    Workflow queues allow workflows to be started at a later time, based on concurrency and
    rate limits.
    """

    def __init__(
        self,
        name: str,
        concurrency: Optional[int] = None,
        limiter: Optional[QueueRateLimit] = None,
        *,  # Disable positional arguments from here on
        worker_concurrency: Optional[int] = None,
        priority_enabled: bool = False,
        partition_queue: bool = False,
        polling_interval_sec: float = 1.0,
    ) -> None:
        if (
            worker_concurrency is not None
            and concurrency is not None
            and worker_concurrency > concurrency
        ):
            raise ValueError(
                "worker_concurrency must be less than or equal to concurrency"
            )
        if polling_interval_sec <= 0.0:
            raise ValueError("polling_interval_sec must be positive")
        self.name = name
        self.concurrency = concurrency
        self.worker_concurrency = worker_concurrency
        self.limiter = limiter
        self.priority_enabled = priority_enabled
        self.partition_queue = partition_queue
        self.polling_interval_sec = polling_interval_sec
        from ._dbos import _get_or_create_dbos_registry

        registry = _get_or_create_dbos_registry()
        if self.name in registry.queue_info_map and self.name != INTERNAL_QUEUE_NAME:
            raise Exception(f"Queue {name} has already been declared")
        registry.queue_info_map[self.name] = self

    def enqueue(
        self, func: "Callable[P, R]", *args: P.args, **kwargs: P.kwargs
    ) -> "WorkflowHandle[R]":
        from ._dbos import _get_dbos_instance

        context = get_local_dbos_context()
        if (
            context is not None
            and context.priority is not None
            and not self.priority_enabled
        ):
            raise Exception(
                f"Priority is not enabled for queue {self.name}. Setting priority will not have any effect."
            )
        if self.partition_queue and (
            context is None or context.queue_partition_key is None
        ):
            raise Exception(
                f"A workflow cannot be enqueued on partitioned queue {self.name} without a partition key"
            )
        if context and context.queue_partition_key and not self.partition_queue:
            raise Exception(
                f"You can only use a partition key on a partition-enabled queue. Key {context.queue_partition_key} was used with non-partitioned queue {self.name}"
            )
        if context and context.queue_partition_key and context.deduplication_id:
            raise Exception("Deduplication is not supported for partitioned queues")

        dbos = _get_dbos_instance()
        return start_workflow(
            dbos, func, args, kwargs, queue_name=self.name, execute_workflow=False
        )

    async def enqueue_async(
        self,
        func: "Callable[P, Coroutine[Any, Any, R]]",
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> "WorkflowHandleAsync[R]":
        from ._dbos import _get_dbos_instance

        dbos = _get_dbos_instance()
        ctx = get_local_dbos_context()
        parent_ctx_copy = copy.copy(ctx)
        child_ctx = DBOSContext.create_start_workflow_child(ctx)
        return await start_workflow_async(
            dbos,
            parent_ctx_copy,
            child_ctx,
            func,
            args,
            kwargs,
            queue_name=self.name,
            execute_workflow=False,
        )


def queue_worker_thread(
    stop_event: threading.Event, dbos: "DBOS", queue: Queue
) -> None:
    """Worker thread for processing a single queue."""
    polling_interval = queue.polling_interval_sec
    min_polling_interval = queue.polling_interval_sec
    max_polling_interval = max(queue.polling_interval_sec, 120.0)

    while not stop_event.is_set():
        # Wait for the polling interval with jitter
        if stop_event.wait(timeout=polling_interval * random.uniform(0.95, 1.05)):
            return

        try:
            if queue.partition_queue:
                queue_partition_keys = dbos._sys_db.get_queue_partitions(queue.name)
                for key in queue_partition_keys:
                    dequeued_workflows = dbos._sys_db.start_queued_workflows(
                        queue,
                        GlobalParams.executor_id,
                        GlobalParams.app_version,
                        key,
                    )
                    for id in dequeued_workflows:
                        try:
                            execute_workflow_by_id(dbos, id, False, True)
                        except Exception as e:
                            dbos.logger.error(f"Error executing workflow {id}: {e}")
            else:
                dequeued_workflows = dbos._sys_db.start_queued_workflows(
                    queue, GlobalParams.executor_id, GlobalParams.app_version, None
                )
                for id in dequeued_workflows:
                    try:
                        execute_workflow_by_id(dbos, id, False, True)
                    except Exception as e:
                        dbos.logger.error(f"Error executing workflow {id}: {e}")
        except OperationalError as e:
            if isinstance(
                e.orig, (errors.SerializationFailure, errors.LockNotAvailable)
            ):
                # If a serialization error is encountered, increase the polling interval
                polling_interval = min(
                    max_polling_interval,
                    polling_interval * 2.0,
                )
                dbos.logger.warning(
                    f"Contention detected in queue thread for {queue.name}. Increasing polling interval to {polling_interval:.2f}."
                )
            else:
                dbos.logger.warning(
                    f"Exception encountered in queue thread for {queue.name}: {e}"
                )
        except Exception as e:
            if not stop_event.is_set():
                # Only print the error if the thread is not stopping
                dbos.logger.warning(
                    f"Exception encountered in queue thread for {queue.name}: {e}"
                )

        # Attempt to scale back the polling interval on each iteration
        polling_interval = max(min_polling_interval, polling_interval * 0.9)


def queue_thread(stop_event: threading.Event, dbos: "DBOS") -> None:
    """Main queue manager thread that spawns and monitors worker threads for each queue."""
    queue_threads: dict[str, threading.Thread] = {}
    # Check interval for monitoring queue registration changes
    check_interval = 1.0

    if dbos._listening_queues is not None:
        listening_queues = dbos._listening_queues
    else:
        listening_queues = list(dbos._registry.queue_info_map.values())
        listening_queues = [
            q for q in listening_queues if q.name != INTERNAL_QUEUE_NAME
        ]
    dbos.logger.info(f"Listening to {len(listening_queues)} queues:")
    log_queues(listening_queues)

    while not stop_event.is_set():
        if dbos._listening_queues is not None:
            # If explicitly listening for queues, only use those queues
            current_queues = {queue.name: queue for queue in dbos._listening_queues}
            # Always listen to the internal queue
            current_queues[INTERNAL_QUEUE_NAME] = dbos._registry.get_internal_queue()
        else:
            # Else, check all declared queues
            current_queues = dict(dbos._registry.queue_info_map)

        # Start threads for new queues
        for queue_name, queue in current_queues.items():
            if (
                queue_name not in queue_threads
                or not queue_threads[queue_name].is_alive()
            ):
                thread = threading.Thread(
                    target=queue_worker_thread,
                    args=(stop_event, dbos, queue),
                    name=f"queue-worker-{queue_name}",
                    daemon=True,
                )
                thread.start()
                queue_threads[queue_name] = thread
                dbos.logger.debug(f"Started worker thread for queue: {queue_name}")

        # Wait for the check interval or stop event
        if stop_event.wait(timeout=check_interval):
            break

    # Join all queue worker threads
    dbos.logger.info("Stopping queue manager, joining all worker threads...")
    for queue_name, thread in queue_threads.items():
        if thread.is_alive():
            thread.join(timeout=10.0)  # Give each thread 10 seconds to finish
            if thread.is_alive():
                dbos.logger.debug(
                    f"Queue worker thread for {queue_name} did not stop in time"
                )
            else:
                dbos.logger.debug(
                    f"Queue worker thread for {queue_name} stopped successfully"
                )


def log_queues(queues: list[Queue]) -> None:
    """Helper function to log queues on DBOS launch."""
    for q in queues:
        opts = []
        if q.concurrency is not None:
            opts.append(f"concurrency={q.concurrency}")
        if q.worker_concurrency is not None:
            opts.append(f"worker_concurrency={q.worker_concurrency}")
        if q.limiter is not None:
            opts.append(f"limit={q.limiter['limit']}/{q.limiter['period']}s")
        if q.priority_enabled:
            opts.append("priority")
        if q.partition_queue:
            opts.append("partitioned")
        opts_str = f" ({', '.join(opts)})" if opts else ""
        dbos_logger.info(f"Queue: {q.name}{opts_str}")
